"""Evaluation route handlers."""
from __future__ import annotations

import os
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ...schemas import (
    EvaluationScorecardResponse,
    EvaluationRegressionDatasetResponse,
    EvaluationJudgeDisagreementsRequest,
    EvaluationJudgeDisagreementsResponse,
    EvaluationContinuousJobCreateRequest,
    EvaluationContinuousJobsResponse,
    EvaluationContinuousIncidentsResponse,
    EvaluationContinuousTrendDeltaResponse,
    EvaluationContinuousIncidentSnapshotResponse,
    EvaluationContinuousIncidentResponse,
    EvaluationContinuousJobResponse,
    EvaluationContinuousJobRunResponse,
    EvaluationChampionChallengerRequest,
    EvaluationChampionChallengerResponse,
)

if TYPE_CHECKING:
    pass


def _build_eval_quality_manager(continuous_eval_path: str | None) -> Any:
    from logos_adk.quality_history import (
        PostgreSQLQualityHistoryStore,
        QualityExecutionManager,
        SQLiteQualityHistoryStore,
    )

    dsn = os.getenv("LOGOS_ADK_QUALITY_DSN")
    explicit_path = os.getenv("LOGOS_ADK_QUALITY_DB_PATH")
    if dsn:
        return QualityExecutionManager(PostgreSQLQualityHistoryStore(dsn=dsn))

    if explicit_path:
        quality_path = explicit_path
    elif continuous_eval_path:
        quality_path = str(Path(continuous_eval_path).with_name("quality_runs.db"))
    else:
        quality_path = ".logos_adk/quality_runs.db"
    return QualityExecutionManager(SQLiteQualityHistoryStore(path=quality_path))


def _get_eval_engine(learning_store: Any | None) -> Any | None:
    if learning_store is None:
        return None
    from logos_adk.learning import EvaluationLoopEngine

    return EvaluationLoopEngine(learning_store)


def _get_eval_runner(
    learning_store: Any | None,
    *,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    restore_jobs: bool = True,
) -> Any | None:
    engine = _get_eval_engine(learning_store)
    if engine is None:
        return None

    from logos_adk.learning import ContinuousEvalRunner

    continuous_eval_path = os.getenv("LOGOS_ADK_CONTINUOUS_EVAL_DB_PATH")
    runner = ContinuousEvalRunner(
        engine=engine,
        quality_manager=_build_eval_quality_manager(continuous_eval_path),
        path=continuous_eval_path,
        agent=agent,
        agent_name=getattr(agent, "name", None),
        prompt_registry=prompt_registry,
    )
    if restore_jobs:
        runner.default_agent_name = None
        runner.restore_jobs()
    return runner


def _build_mock_judge(agent: Any, *, mock_response: str, pass_threshold: float) -> Any:
    from logos_adk.providers.mock import MockProvider
    from logos_adk.testing_framework import LLMJudge

    return LLMJudge(
        agent.__class__("judge", model=MockProvider(responses=[mock_response])),
        pass_threshold=pass_threshold,
    )


async def get_scorecard_handler(
    workspace_id: str | None = None,
    agent_name: str | None = None,
    window_days: int = 7,
    learning_store: Any | None = None,
) -> EvaluationScorecardResponse:
    """Handle GET /learning/evaluation/scorecard."""
    engine = _get_eval_engine(learning_store)
    if engine is None:
        return EvaluationScorecardResponse(scorecard={"error": "learning_store not configured"})

    scorecard = engine.build_scorecard(
        workspace_id=workspace_id,
        agent_name=agent_name,
        window_days=window_days,
    )
    return EvaluationScorecardResponse(scorecard=asdict(scorecard))


async def get_regression_dataset_handler(
    workspace_id: str | None = None,
    agent_name: str | None = None,
    limit: int = 100,
    learning_store: Any | None = None,
) -> EvaluationRegressionDatasetResponse:
    """Handle GET /learning/evaluation/regression-dataset."""
    engine = _get_eval_engine(learning_store)
    if engine is None:
        return EvaluationRegressionDatasetResponse(dataset={"error": "learning_store not configured"})

    dataset = engine.build_regression_dataset(
        workspace_id=workspace_id,
        agent_name=agent_name,
        limit=limit,
    )
    return EvaluationRegressionDatasetResponse(dataset=asdict(dataset))


async def get_judge_disagreements_handler(
    request: EvaluationJudgeDisagreementsRequest,
    agent: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationJudgeDisagreementsResponse:
    """Handle POST /learning/evaluation/judge-disagreements."""
    engine = _get_eval_engine(learning_store)
    if engine is None or agent is None:
        return EvaluationJudgeDisagreementsResponse(
            summary={"error": "learning_store not configured"},
            items=[],
        )

    result = await engine.run_judge_snapshot(
        agent,
        judge=_build_mock_judge(
            agent,
            mock_response=request.mock_response,
            pass_threshold=request.pass_threshold,
        ),
        workspace_id=request.workspace_id,
        agent_name=request.agent_name,
        limit=request.limit,
    )
    return EvaluationJudgeDisagreementsResponse(
        summary=result.get("summary", {}),
        items=result.get("disagreements", []),
    )


async def create_eval_job_handler(
    request: EvaluationContinuousJobCreateRequest,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousJobResponse:
    """Handle POST /learning/evaluation/jobs."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None or agent is None:
        return EvaluationContinuousJobResponse(job={"error": "learning_store not configured"})

    existing = runner.get_job(request.name)
    if existing is not None:
        return EvaluationContinuousJobResponse(job=runner.job_snapshot(existing))

    metadata = dict(request.metadata or {})
    if request.kind == "judge-eval":
        metadata.setdefault("judge_kind", "mock")
        metadata.setdefault("mock_response", request.mock_response)
        metadata.setdefault("pass_threshold", request.pass_threshold)
        job = runner.register_judge_job(
            request.name,
            agent=agent,
            judge=_build_mock_judge(
                agent,
                mock_response=request.mock_response,
                pass_threshold=request.pass_threshold,
            ),
            workspace_id=request.workspace_id,
            agent_name=request.agent_name,
            interval_seconds=request.interval_seconds,
            limit=request.limit,
            metadata=metadata,
        )
    else:
        job = runner.register_regression_job(
            request.name,
            agent=agent,
            workspace_id=request.workspace_id,
            agent_name=request.agent_name,
            interval_seconds=request.interval_seconds,
            limit=request.limit,
            metadata=metadata,
        )
    return EvaluationContinuousJobResponse(job=runner.job_snapshot(job))


async def get_eval_jobs_handler(
    workspace_id: str | None = None,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousJobsResponse:
    """Handle GET /learning/evaluation/jobs."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousJobsResponse(items=[])

    items = runner.list_jobs()
    if workspace_id is not None:
        items = [item for item in items if item.get("workspace_id") == workspace_id]
    return EvaluationContinuousJobsResponse(items=items)


async def get_restored_eval_jobs_handler(
    workspace_id: str | None = None,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousJobsResponse:
    """Handle GET /learning/evaluation/restored-jobs."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousJobsResponse(items=[])

    items = runner.list_restored_jobs()
    if workspace_id is not None:
        items = [item for item in items if item.get("workspace_id") == workspace_id]
    return EvaluationContinuousJobsResponse(items=items)


async def get_eval_incidents_handler(
    workspace_id: str | None = None,
    job_name: str | None = None,
    kind: str | None = None,
    status: str | None = None,
    limit: int = 50,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousIncidentsResponse:
    """Handle GET /learning/evaluation/incidents."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousIncidentsResponse(items=[])

    incidents = runner.list_incidents(
        limit=limit,
        kind=kind,
        job_name=job_name,
        workspace_id=workspace_id,
        status=status,
    )
    return EvaluationContinuousIncidentsResponse(items=incidents)


async def get_eval_trend_handler(
    workspace_id: str | None = None,
    job_name: str | None = None,
    window_days: int = 7,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousTrendDeltaResponse:
    """Handle GET /learning/evaluation/trend-delta."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousTrendDeltaResponse(report={"error": "learning_store not configured"})

    report = runner.trend_delta_report(
        workspace_id=workspace_id,
        job_name=job_name,
        window_days=window_days,
    )
    return EvaluationContinuousTrendDeltaResponse(report=asdict(report))


async def get_incident_snapshot_handler(
    incident_id: str | None = None,
    workspace_id: str | None = None,
    job_name: str | None = None,
    status: str | None = None,
    limit: int = 20,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousIncidentSnapshotResponse:
    """Handle GET /learning/evaluation/incidents/{incident_id}/snapshot."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousIncidentSnapshotResponse(snapshot={"error": "learning_store not configured"})

    snapshot = asdict(
        runner.incident_snapshot(
            workspace_id=workspace_id,
            job_name=job_name,
            status=status,
            limit=limit,
        )
    )
    if incident_id is not None:
        snapshot["incident"] = runner.get_incident(incident_id)
    return EvaluationContinuousIncidentSnapshotResponse(snapshot=snapshot)


async def acknowledge_incident_handler(
    incident_id: str,
    actor: str | None = None,
    note: str | None = None,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousIncidentResponse:
    """Handle POST /learning/evaluation/incidents/{incident_id}/acknowledge."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousIncidentResponse(incident={"error": "learning_store not configured"})

    incident = runner.acknowledge_incident(incident_id, actor=actor, note=note)
    return EvaluationContinuousIncidentResponse(
        incident=incident or {"id": incident_id, "error": "incident not found"}
    )


async def resolve_incident_handler(
    incident_id: str,
    actor: str | None = None,
    note: str | None = None,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousIncidentResponse:
    """Handle POST /learning/evaluation/incidents/{incident_id}/resolve."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousIncidentResponse(incident={"error": "learning_store not configured"})

    incident = runner.resolve_incident(incident_id, actor=actor, note=note)
    return EvaluationContinuousIncidentResponse(
        incident=incident or {"id": incident_id, "error": "incident not found"}
    )


async def get_eval_job_handler(
    job_id: str,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousJobResponse:
    """Handle GET /learning/evaluation/jobs/{job_id}."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousJobResponse(job={"error": "learning_store not configured"})

    job = runner.get_job(job_id)
    if job is None:
        return EvaluationContinuousJobResponse(job={"error": "job not found"})
    return EvaluationContinuousJobResponse(job=runner.job_snapshot(job))


async def run_eval_job_handler(
    job_id: str,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousJobRunResponse:
    """Handle POST /learning/evaluation/jobs/{job_id}/run."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousJobRunResponse(
            job={"error": "learning_store not configured"},
            run={},
        )

    if runner.get_job(job_id) is None:
        return EvaluationContinuousJobRunResponse(job={"error": "job not found"}, run={})

    run = await runner.run_job_once(job_id)
    job = runner.get_job(job_id)
    return EvaluationContinuousJobRunResponse(
        job=runner.job_snapshot(job) if job is not None else {"id": job_id},
        run=asdict(run),
    )


async def delete_eval_job_handler(
    job_id: str,
    agent: Any | None = None,
    prompt_registry: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationContinuousJobResponse:
    """Handle DELETE /learning/evaluation/jobs/{job_id}."""
    runner = _get_eval_runner(
        learning_store,
        agent=agent,
        prompt_registry=prompt_registry,
    )
    if runner is None:
        return EvaluationContinuousJobResponse(job={"error": "learning_store not configured"})

    removed = await runner.remove_job(job_id)
    return EvaluationContinuousJobResponse(job=removed or {"error": "job not found"})


async def champion_challenger_handler(
    request: EvaluationChampionChallengerRequest,
    agent: Any | None = None,
    learning_store: Any | None = None,
) -> EvaluationChampionChallengerResponse:
    """Handle POST /learning/evaluation/champion-challenger."""
    engine = _get_eval_engine(learning_store)
    if engine is None or agent is None:
        return EvaluationChampionChallengerResponse(report={"error": "learning_store not configured"})

    report = await engine.compare_champion_challenger(
        agent,
        challenger_prompt=request.challenger_prompt,
        champion_prompt=request.champion_prompt,
        workspace_id=request.workspace_id,
        agent_name=request.agent_name,
        champion_filter=request.champion_filter,
        challenger_filter=request.challenger_filter,
        limit=request.limit,
        reward_scorer_mode=request.reward_scorer_mode,
        reward_training_limit=request.reward_training_limit,
    )
    return EvaluationChampionChallengerResponse(report=asdict(report))


__all__ = [
    "get_scorecard_handler",
    "get_regression_dataset_handler",
    "get_judge_disagreements_handler",
    "create_eval_job_handler",
    "get_eval_jobs_handler",
    "get_restored_eval_jobs_handler",
    "get_eval_incidents_handler",
    "get_eval_trend_handler",
    "get_incident_snapshot_handler",
    "acknowledge_incident_handler",
    "resolve_incident_handler",
    "get_eval_job_handler",
    "run_eval_job_handler",
    "delete_eval_job_handler",
    "champion_challenger_handler",
]



"""Learning CLI groups - learning, retrieval, tool-workflow, model-routing, evaluation, improvement, transfer."""
from __future__ import annotations

import asyncio
import json
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click

from ..utils.output import write_result_output

if TYPE_CHECKING:
    pass


def _retrieval_dsn_option(f: Any) -> Any:
    """Add --retrieval-dsn option to command."""
    return click.option(
        "--retrieval-dsn",
        help="Override retrieval learning DSN (PostgreSQL)",
    )(f)


def _build_retrieval_engine(
    db_path: str | None,
    config_path: Path | None,
    dsn: str | None = None,
) -> Any:
    """Build retrieval learning engine."""
    from logos_adk.learning.retrieval_learning import RetrievalLearningEngine
    
    if dsn:
        return RetrievalLearningEngine(dsn=dsn)
    
    if config_path:
        from logos_adk.config_loader import load_agent_config
        config = load_agent_config(str(config_path))
        return RetrievalLearningEngine.from_config(config)
    
    return RetrievalLearningEngine(path=db_path or ".logos_adk/retrieval_learning.db")


def _write_and_echo_retrieval_items(
    *,
    output: Path | None,
    payload: dict[str, Any],
    empty_message: str,
    render: Any,
) -> None:
    """Write payload to file and print a human-readable summary."""
    write_result_output(output, payload)
    items = payload["items"]
    if not items:
        click.echo(empty_message)
        return
    render(items)


def _require_retrieval_request(action: str, request: Any | None, request_id: str) -> Any:
    """Raise a ClickException when a reindex request action cannot be applied."""
    if request is None:
        raise click.ClickException(f"Unknown or unsupported reindex request for {action}: {request_id}")
    return request


def _build_prompt_registry(path: str | None) -> Any | None:
    """Build prompt registry when a path is provided."""
    if not path:
        return None
    from logos_adk.learning import PromptRegistry

    return PromptRegistry(path)


def _build_evaluation_components(
    *,
    learning_db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    agent: Any | None = None,
) -> tuple[Any, Any, Any]:
    """Build runner, engine, and prompt registry for evaluation CLI commands."""
    from logos_adk.learning import EvaluationLoopEngine

    from ..helpers import (
        _build_continuous_eval_runner,
        _build_learning_store,
        _build_quality_manager,
    )

    learning_store = _build_learning_store(learning_db_path)
    quality_manager = _build_quality_manager(quality_db_path)
    engine = EvaluationLoopEngine(learning_store)
    prompt_registry = _build_prompt_registry(prompt_registry_path)
    runner = _build_continuous_eval_runner(
        learning_store,
        quality_manager,
        engine,
        path=continuous_eval_db_path,
        agent=agent,
        prompt_registry=prompt_registry,
        restore_jobs=True,
        restore_all_jobs=True,
    )
    return runner, engine, prompt_registry


def _load_agent_from_config(agent_config: Path | None) -> Any | None:
    """Load an agent from config when the CLI command needs one."""
    if agent_config is None:
        return None
    from logos_adk.cli import Agent

    return Agent.from_config(str(agent_config))


def _parse_json_option(value: str | None, *, option_name: str) -> dict[str, Any] | None:
    """Parse a JSON-encoded CLI option."""
    if value is None:
        return None
    try:
        parsed = json.loads(value)
    except json.JSONDecodeError as exc:
        raise click.ClickException(f"Invalid JSON for {option_name}: {exc}") from exc
    if parsed is not None and not isinstance(parsed, dict):
        raise click.ClickException(f"{option_name} must be a JSON object")
    return parsed


# =============================================================================
# Storage Group
# =============================================================================

@click.group()
def storage() -> None:
    """Persistence schema and migration commands."""
    pass


@storage.command("upgrade-check")
@click.option("--kind", type=click.Choice(["state", "memory", "cache", "quality"]), required=True)
@click.option("--backend", type=click.Choice(["sqlite", "postgresql"]), default="sqlite")
@click.option("--path", help="SQLite database path")
@click.option("--dsn", help="PostgreSQL DSN")
def storage_upgrade_check(kind: str, backend: str, path: str | None, dsn: str | None) -> None:
    """Check whether a persistence target needs migration."""
    if backend == "sqlite":
        sqlite_path = path or {
            "state": ".logos_adk/state.db",
            "memory": ".logos_adk/memory.db",
            "cache": ".logos_cache.db",
            "quality": ".logos_adk/quality_runs.db",
        }[kind]
        
        from logos_adk.persistence import get_sqlite_schema_version
        import sqlite3
        
        connection = sqlite3.connect(sqlite_path)
        try:
            connection.row_factory = sqlite3.Row
            current = get_sqlite_schema_version(connection, {
                "state": "state_store",
                "memory": "memory_backend",
                "cache": "cache_backend",
                "quality": "quality_history",
            }[kind]) or 0
        finally:
            connection.close()
        
        click.echo(f"{kind}:{backend}")
        click.echo(f"  path: {sqlite_path}")
        click.echo(f"  schema_version: {current}")
        click.echo("  expected_version: 1")
        click.echo(f"  upgrade_required: {'yes' if current < 1 else 'no'}")
        return
    
    from logos_adk.persistence import get_postgresql_schema_version
    from logos_adk.postgres_utils import import_psycopg, resolve_database_url
    
    resolved_dsn = resolve_database_url(dsn)
    if not resolved_dsn:
        raise click.ClickException("PostgreSQL upgrade-check requires a DSN or DATABASE_URL")
    
    psycopg = import_psycopg()
    with psycopg.connect(resolved_dsn) as connection:
        current = get_postgresql_schema_version(connection, {
            "state": "state_store",
            "memory": "memory_backend",
            "cache": "cache_backend",
            "quality": "quality_history",
        }[kind]) or 0
    
    click.echo(f"{kind}:{backend}")
    click.echo(f"  dsn: {resolved_dsn}")
    click.echo(f"  schema_version: {current}")
    click.echo("  expected_version: 1")
    click.echo(f"  upgrade_required: {'yes' if current < 1 else 'no'}")


@storage.command("migrate")
@click.option("--kind", type=click.Choice(["state", "memory", "cache", "quality"]), required=True)
@click.option("--backend", type=click.Choice(["sqlite", "postgresql"]), default="sqlite")
@click.option("--path", help="SQLite database path")
@click.option("--dsn", help="PostgreSQL DSN")
def storage_migrate(kind: str, backend: str, path: str | None, dsn: str | None) -> None:
    """Create/upgrade storage schema metadata and run built-in migrations."""
    if kind == "state" and backend == "sqlite":
        from logos_adk.state_store import SQLiteStateStore
        store = SQLiteStateStore(path or ".logos_adk/state.db")
        with store._connect() as connection:
            store._ensure_schema(connection)
    elif kind == "state" and backend == "postgresql":
        from logos_adk.state_store import PostgreSQLStateStore
        store = PostgreSQLStateStore(dsn or "")
        with store._connect() as connection:
            store._ensure_schema(connection)
    elif kind == "memory" and backend == "sqlite":
        from logos_adk.memory.backends.sqlite import SQLiteBackend
        backend_obj = SQLiteBackend(path or ".logos_adk/memory.db")
        with backend_obj._connect() as connection:
            backend_obj._ensure_schema(connection)
    elif kind == "memory" and backend == "postgresql":
        from logos_adk.memory.backends.postgresql import PostgreSQLBackend
        backend_obj = PostgreSQLBackend(dsn or "")
        with backend_obj._connect() as connection:
            backend_obj._ensure_schema(connection)
    elif kind == "cache" and backend == "sqlite":
        from logos_adk.cache.sqlite import SQLiteCacheBackend
        SQLiteCacheBackend(path or ".logos_cache.db")
    elif kind == "cache" and backend == "postgresql":
        from logos_adk.cache.postgresql import PostgreSQLCacheBackend
        backend_obj = PostgreSQLCacheBackend(dsn=dsn)
        with backend_obj._connect() as connection:
            backend_obj._ensure_schema(connection)
    elif kind == "quality" and backend == "sqlite":
        from logos_adk.quality_history import SQLiteQualityHistoryStore
        store = SQLiteQualityHistoryStore(path=path or ".logos_adk/quality_runs.db")
        with store._connect() as connection:
            store._ensure_schema(connection)
    elif kind == "quality" and backend == "postgresql":
        from logos_adk.quality_history import PostgreSQLQualityHistoryStore
        store = PostgreSQLQualityHistoryStore(dsn or "")
        with store._connect() as connection:
            store._ensure_schema(connection)
    else:
        raise click.ClickException(f"Unsupported storage target: {kind}/{backend}")
    
    click.echo(f"Migration complete for {kind}:{backend}")


# =============================================================================
# Retrieval Group
# =============================================================================

@click.group()
def retrieval() -> None:
    """Inspect retrieval-learning state for operators."""
    pass


@retrieval.command("source-scores")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_source_scores(retrieval_dsn: str | None, db_path: str | None, config_path: Path | None, workspace_id: str | None, output: Path | None) -> None:
    """Print learned source usefulness scores."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    items = [asdict(item) for item in engine.list_source_scores(workspace_id=workspace_id)]
    payload = {"workspace_id": workspace_id, "items": items}
    write_result_output(output, payload)
    
    if not items:
        click.echo("No source scores found.")
        return
    
    click.echo(f"Workspace: {workspace_id or '*'}")
    for item in items:
        click.echo(
            f"{item['source_key']}: score={item['usefulness_score']} "
            f"helpful={item['helpful_count']} harmful={item['harmful_count']} "
            f"accepted={item['accepted_count']} rejected={item['rejected_count']}"
        )
        if item.get("connector"):
            click.echo(f"  connector: {item['connector']}")


@retrieval.command("reindex-triggers")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_reindex_triggers(retrieval_dsn: str | None, db_path: str | None, config_path: Path | None, workspace_id: str | None, output: Path | None) -> None:
    """Print open retrieval reindex triggers."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    items = [asdict(item) for item in engine.list_reindex_triggers(workspace_id=workspace_id)]
    payload = {"workspace_id": workspace_id, "items": items}
    write_result_output(output, payload)
    
    if not items:
        click.echo("No reindex triggers found.")
        return
    
    click.echo(f"Workspace: {workspace_id or '*'}")
    for item in items:
        click.echo(
            f"{item['source_key']}: reason={item['reason']} severity={item['severity']} "
            f"occurrences={item['occurrences']} chunk_size={item['suggested_chunk_size']} "
            f"overlap={item['suggested_overlap']}"
        )


@retrieval.command("profiles")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--task-type", help="Task type segment")
@click.option("--user-segment", help="User segment")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_profiles(
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    workspace_id: str | None,
    task_type: str | None,
    user_segment: str | None,
    output: Path | None,
) -> None:
    """Print learned retrieval profiles."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    items = [
        asdict(item)
        for item in engine.list_profiles(
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
        )
    ]
    payload = {
        "workspace_id": workspace_id,
        "task_type": task_type,
        "user_segment": user_segment,
        "items": items,
    }
    write_result_output(output, payload)
    
    if not items:
        click.echo("No retrieval profiles found.")
        return
    
    click.echo(f"Workspace: {workspace_id or '*'} task_type: {task_type or '*'} user_segment: {user_segment or '*'}")
    for item in items:
        click.echo(
            f"limit={item['limit']} score={item['score']} success={item['success_count']} "
            f"failure={item['failure_count']} filter={json.dumps(item['filter'], sort_keys=True)}"
        )


@retrieval.command("ack-trigger")
@click.argument("trigger_id")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--note", help="Operator note")
def retrieval_ack_trigger(trigger_id: str, retrieval_dsn: str | None, db_path: str | None, config_path: Path | None, note: str | None) -> None:
    """Acknowledge a reindex trigger."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    trigger = engine.acknowledge_reindex_trigger(trigger_id, note=note)
    if trigger is None:
        raise click.ClickException(f"Unknown reindex trigger: {trigger_id}")
    click.echo(f"{trigger.id}: status={trigger.status} source={trigger.source_key}")


@retrieval.command("resolve-trigger")
@click.argument("trigger_id")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--note", help="Operator note")
def retrieval_resolve_trigger(trigger_id: str, retrieval_dsn: str | None, db_path: str | None, config_path: Path | None, note: str | None) -> None:
    """Resolve a reindex trigger."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    trigger = engine.resolve_reindex_trigger(trigger_id, note=note)
    if trigger is None:
        raise click.ClickException(f"Unknown reindex trigger: {trigger_id}")
    click.echo(f"{trigger.id}: status={trigger.status} source={trigger.source_key}")


@retrieval.command("close-trigger")
@click.argument("trigger_id")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--note", help="Operator note")
def retrieval_close_trigger(trigger_id: str, retrieval_dsn: str | None, db_path: str | None, config_path: Path | None, note: str | None) -> None:
    """Close a reindex trigger."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    trigger = engine.close_reindex_trigger(trigger_id, note=note)
    if trigger is None:
        raise click.ClickException(f"Unknown reindex trigger: {trigger_id}")
    click.echo(f"{trigger.id}: status={trigger.status} source={trigger.source_key}")


@retrieval.command("source-documents")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--source-key", help="Source identifier")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_source_documents(
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    workspace_id: str | None,
    source_key: str | None,
    output: Path | None,
) -> None:
    """Print stored source-document snapshots."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    items = []
    for item in engine.list_source_documents(workspace_id=workspace_id, source_key=source_key):
        payload = asdict(item)
        payload["content_length"] = len(item.content)
        items.append(payload)
    _write_and_echo_retrieval_items(
        output=output,
        payload={"workspace_id": workspace_id, "source_key": source_key, "items": items},
        empty_message="No source documents found.",
        render=lambda rows: (
            click.echo(f"Workspace: {workspace_id or '*'} source_key: {source_key or '*'}"),
            [click.echo(
                f"{row['source_key']}:{row['document_id']} length={row['content_length']} updated={row['updated_at']}"
            ) for row in rows],
        ),
    )


@retrieval.command("force-reindex")
@click.argument("source_key")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--reason", default="operator_requested", help="Reason for the request")
@click.option("--requested-by", help="Operator or system requesting reindex")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_force_reindex(
    source_key: str,
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    workspace_id: str | None,
    reason: str,
    requested_by: str | None,
    output: Path | None,
) -> None:
    """Create a manual reindex request."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    request = engine.create_force_reindex_request(
        workspace_id=workspace_id,
        source_key=source_key,
        reason=reason,
        requested_by=requested_by,
    )
    payload = {"request": asdict(request)}
    write_result_output(output, payload)
    click.echo(f"{request.id}: status={request.status} source={request.source_key}")


@retrieval.command("export-source-scores")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--format", "output_format", type=click.Choice(["json", "csv"]), default="json")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_export_source_scores(
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    workspace_id: str | None,
    output_format: str,
    output: Path | None,
) -> None:
    """Export learned source usefulness scores."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    report = engine.export_source_scores_report(workspace_id=workspace_id, format=output_format)
    if output is not None:
        output.parent.mkdir(parents=True, exist_ok=True)
        output.write_text(report)
    click.echo(report)


@retrieval.command("reindex-requests")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_reindex_requests(
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    workspace_id: str | None,
    output: Path | None,
) -> None:
    """Print reindex requests."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    items = [asdict(item) for item in engine.list_reindex_requests(workspace_id=workspace_id)]
    _write_and_echo_retrieval_items(
        output=output,
        payload={"workspace_id": workspace_id, "items": items},
        empty_message="No reindex requests found.",
        render=lambda rows: (
            click.echo(f"Workspace: {workspace_id or '*'}"),
            [click.echo(
                f"{row['id']}: status={row['status']} source={row['source_key']} requested_by={row['requested_by']}"
            ) for row in rows],
        ),
    )


@retrieval.command("cancel-request")
@click.argument("request_id")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--note", help="Operator note")
def retrieval_cancel_request(
    request_id: str,
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    note: str | None,
) -> None:
    """Cancel a queued or running reindex request."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    request = _require_retrieval_request(
        "cancel",
        engine.cancel_reindex_request(request_id, note=note),
        request_id,
    )
    click.echo(f"{request.id}: status={request.status} source={request.source_key}")


@retrieval.command("resume-request")
@click.argument("request_id")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--note", help="Operator note")
def retrieval_resume_request(
    request_id: str,
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    note: str | None,
) -> None:
    """Resume a canceled reindex request."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    request = _require_retrieval_request(
        "resume",
        engine.resume_reindex_request(request_id, note=note),
        request_id,
    )
    click.echo(f"{request.id}: status={request.status} source={request.source_key}")


@retrieval.command("replay-request")
@click.argument("request_id")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--note", help="Operator note")
def retrieval_replay_request(
    request_id: str,
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    note: str | None,
) -> None:
    """Replay a canceled reindex request as a fresh request."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    request = _require_retrieval_request(
        "replay",
        engine.replay_reindex_request(request_id, note=note),
        request_id,
    )
    click.echo(f"{request.id}: status={request.status} source={request.source_key}")


@retrieval.command("summary")
@_retrieval_dsn_option
@click.option("--db-path", default=None, help="Retrieval learning SQLite DB path")
@click.option("--config", "config_path", type=click.Path(exists=True, path_type=Path), help="Agent YAML config")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def retrieval_summary(
    retrieval_dsn: str | None,
    db_path: str | None,
    config_path: Path | None,
    workspace_id: str | None,
    output: Path | None,
) -> None:
    """Print retrieval operations summary."""
    engine = _build_retrieval_engine(db_path, config_path, dsn=retrieval_dsn)
    payload = engine.ops_summary(workspace_id=workspace_id)
    write_result_output(output, payload)
    click.echo(f"Workspace: {workspace_id or '*'}")
    click.echo(f"Requests: {payload['requests']['total']}")
    click.echo(f"Triggers: {payload['triggers']['total']}")
    click.echo(f"Source Documents: {payload['source_documents']['total']}")
    click.echo(f"Coalescing: {payload['coalescing']['coalesced_total']}")


# =============================================================================
# Tool Workflow Group
# =============================================================================

@click.group("tool-workflow")
def tool_workflow() -> None:
    """Inspect tool and workflow learning state for operators."""
    pass


# =============================================================================
# Model Routing Group
# =============================================================================

@click.group("model-routing")
def model_routing() -> None:
    """Inspect model routing learning state for operators."""
    pass


# =============================================================================
# Evaluation Group
# =============================================================================

@click.group()
def evaluation() -> None:
    """Inspect and run evaluation loop workflows."""
    pass


@evaluation.command("scorecard")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--agent-name", help="Agent scope")
@click.option("--window-days", default=7, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_scorecard(
    db_path: str | None,
    workspace_id: str | None,
    agent_name: str | None,
    window_days: int,
    output: Path | None,
) -> None:
    """Print an operator scorecard for recent production behavior."""
    from logos_adk.learning import EvaluationLoopEngine

    from ..helpers import _build_learning_store

    engine = EvaluationLoopEngine(_build_learning_store(db_path))
    scorecard = asdict(
        engine.build_scorecard(
            workspace_id=workspace_id,
            agent_name=agent_name,
            window_days=window_days,
        )
    )
    write_result_output(output, {"scorecard": scorecard})
    click.echo(
        "quality_delta={quality:.4f} retrieval_delta={retrieval:.4f} "
        "tool_delta={tool:.4f} latency_delta={latency:.4f}".format(
            quality=float(scorecard["quality"]["delta"]),
            retrieval=float(scorecard["retrieval_relevance"]["delta"]),
            tool=float(scorecard["tool_success"]["delta"]),
            latency=float(scorecard["latency"]["delta"]),
        )
    )


@evaluation.command("regression-dataset")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--agent-name", help="Agent scope")
@click.option("--limit", default=100, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_regression_dataset(
    db_path: str | None,
    workspace_id: str | None,
    agent_name: str | None,
    limit: int,
    output: Path | None,
) -> None:
    """Print a regression dataset derived from production feedback."""
    from logos_adk.learning import EvaluationLoopEngine

    from ..helpers import _build_learning_store

    engine = EvaluationLoopEngine(_build_learning_store(db_path))
    dataset = asdict(
        engine.build_regression_dataset(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
        )
    )
    write_result_output(output, {"dataset": dataset})
    click.echo(
        "examples={count} workspace_id={workspace} agent_name={agent}".format(
            count=len(dataset["examples"]),
            workspace=dataset.get("workspace_id") or "*",
            agent=dataset.get("agent_name") or "*",
        )
    )


@evaluation.command("judge-disagreements")
@click.argument("agent_config", type=click.Path(exists=True, path_type=Path))
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--agent-name", help="Agent scope")
@click.option("--limit", default=100, type=int, show_default=True)
@click.option("--threshold", default=0.35, type=float, show_default=True)
@click.option("--pass-threshold", default=0.7, type=float, show_default=True)
@click.option("--mock-response", default='{"score": 0.5, "rationale": "no judge configured"}', show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_judge_disagreements(
    agent_config: Path,
    db_path: str | None,
    workspace_id: str | None,
    agent_name: str | None,
    limit: int,
    threshold: float,
    pass_threshold: float,
    mock_response: str,
    output: Path | None,
) -> None:
    """Run a judge snapshot and print human-vs-judge disagreements."""
    from logos_adk.learning import EvaluationLoopEngine
    from logos_adk.providers.mock import MockProvider
    from logos_adk.testing_framework import LLMJudge

    from ..helpers import _build_learning_store

    agent = _load_agent_from_config(agent_config)
    engine = EvaluationLoopEngine(_build_learning_store(db_path))
    judge = LLMJudge(
        agent.__class__("judge", model=MockProvider(responses=[mock_response])),
        pass_threshold=pass_threshold,
    )
    payload = asyncio.run(
        engine.run_judge_snapshot(
            agent,
            judge=judge,
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
        )
    )
    result = {
        "summary": payload.get("summary", {}),
        "items": payload.get("disagreements", []),
    }
    write_result_output(output, result)
    click.echo(
        "disagreements={count} judge_disagreements={summary}".format(
            count=len(result["items"]),
            summary=result["summary"].get("judge_disagreements", 0),
        )
    )


@evaluation.command("champion-challenger")
@click.argument("agent_config", type=click.Path(exists=True, path_type=Path))
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--agent-name", help="Agent scope")
@click.option("--champion-prompt", default=None)
@click.option("--challenger-prompt", required=True)
@click.option("--champion-filter-json", default=None)
@click.option("--challenger-filter-json", default=None)
@click.option("--limit", default=100, type=int, show_default=True)
@click.option("--reward-scorer-mode", default="heuristic", show_default=True)
@click.option("--reward-training-limit", default=500, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_champion_challenger(
    agent_config: Path,
    db_path: str | None,
    workspace_id: str | None,
    agent_name: str | None,
    champion_prompt: str | None,
    challenger_prompt: str,
    champion_filter_json: str | None,
    challenger_filter_json: str | None,
    limit: int,
    reward_scorer_mode: str,
    reward_training_limit: int,
    output: Path | None,
) -> None:
    """Compare champion and challenger prompts against production-derived cases."""
    from logos_adk.learning import EvaluationLoopEngine

    from ..helpers import _build_learning_store

    agent = _load_agent_from_config(agent_config)
    engine = EvaluationLoopEngine(_build_learning_store(db_path))
    report = asdict(
        asyncio.run(
            engine.compare_champion_challenger(
                agent,
                challenger_prompt=challenger_prompt,
                champion_prompt=champion_prompt,
                workspace_id=workspace_id,
                agent_name=agent_name,
                champion_filter=_parse_json_option(
                    champion_filter_json,
                    option_name="--champion-filter-json",
                ),
                challenger_filter=_parse_json_option(
                    challenger_filter_json,
                    option_name="--challenger-filter-json",
                ),
                limit=limit,
                reward_scorer_mode=reward_scorer_mode,
                reward_training_limit=reward_training_limit,
            )
        )
    )
    write_result_output(output, {"report": report})
    click.echo(f"Recommendation: {report['recommendation']}")


@evaluation.command("run-regression-job")
@click.argument("agent_config", type=click.Path(exists=True, path_type=Path))
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--agent-name", help="Agent scope")
@click.option("--name", required=True, help="Job name")
@click.option("--interval-seconds", default=3600.0, type=float, show_default=True)
@click.option("--limit", default=100, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_run_regression_job(
    agent_config: Path,
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    workspace_id: str | None,
    agent_name: str | None,
    name: str,
    interval_seconds: float,
    limit: int,
    output: Path | None,
) -> None:
    """Register and run one regression job immediately."""
    agent = _load_agent_from_config(agent_config)
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
        agent=agent,
    )
    if runner.get_job(name) is None:
        runner.register_regression_job(
            name,
            agent=agent,
            workspace_id=workspace_id,
            agent_name=agent_name,
            interval_seconds=interval_seconds,
            limit=limit,
        )
    run = asyncio.run(runner.run_job_once(name))
    job = runner.get_job(name)
    payload = {
        "job": runner.job_snapshot(job) if job is not None else {"name": name},
        "run": asdict(run),
    }
    write_result_output(output, payload)
    click.echo(f"{name}: status={payload['run']['status']}")


@evaluation.command("restored-jobs")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_restored_jobs(
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    workspace_id: str | None,
    output: Path | None,
) -> None:
    """List restored continuous evaluation jobs."""
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
    )
    items = runner.list_restored_jobs()
    if workspace_id is not None:
        items = [item for item in items if item.get("workspace_id") == workspace_id]
    payload = {"items": items}
    write_result_output(output, payload)
    if not items:
        click.echo("No restored jobs found.")
        return
    for item in items:
        click.echo(f"{item['name']}: kind={item['kind']} restored={item['restored']}")


@evaluation.command("trend-deltas")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--job-name", default=None)
@click.option("--window-days", default=7, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_trend_deltas(
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    workspace_id: str | None,
    job_name: str | None,
    window_days: int,
    output: Path | None,
) -> None:
    """Print trend deltas for recent continuous evaluation activity."""
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
    )
    report = asdict(
        runner.trend_delta_report(
            workspace_id=workspace_id,
            job_name=job_name,
            window_days=window_days,
        )
    )
    write_result_output(output, {"report": report})
    click.echo(f"incidents_delta={report['incidents']['delta']:.4f}")


@evaluation.command("rollback-incidents")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--job-name", default=None)
@click.option("--status", default=None)
@click.option("--limit", default=50, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_rollback_incidents(
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    workspace_id: str | None,
    job_name: str | None,
    status: str | None,
    limit: int,
    output: Path | None,
) -> None:
    """List rollback and regression incidents."""
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
    )
    items = runner.list_incidents(
        limit=limit,
        job_name=job_name,
        workspace_id=workspace_id,
        status=status,
    )
    payload = {"items": items}
    write_result_output(output, payload)
    if not items:
        click.echo("No incidents found.")
        return
    for item in items:
        click.echo(f"{item['id']}: {item['kind']} status={item['status']}")


@evaluation.command("incident-snapshot")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--workspace-id", help="Workspace scope")
@click.option("--job-name", default=None)
@click.option("--status", default=None)
@click.option("--limit", default=20, type=int, show_default=True)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_incident_snapshot(
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    workspace_id: str | None,
    job_name: str | None,
    status: str | None,
    limit: int,
    output: Path | None,
) -> None:
    """Print an aggregate snapshot of recent incidents."""
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
    )
    snapshot = asdict(
        runner.incident_snapshot(
            workspace_id=workspace_id,
            job_name=job_name,
            status=status,
            limit=limit,
        )
    )
    write_result_output(output, {"snapshot": snapshot})
    click.echo(f"total_incidents={snapshot['summary']['total_incidents']}")


@evaluation.command("ack-incident")
@click.argument("incident_id")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--actor", default=None)
@click.option("--note", default=None)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_ack_incident(
    incident_id: str,
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    actor: str | None,
    note: str | None,
    output: Path | None,
) -> None:
    """Acknowledge an evaluation incident."""
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
    )
    incident = runner.acknowledge_incident(incident_id, actor=actor, note=note)
    if incident is None:
        raise click.ClickException(f"Unknown incident: {incident_id}")
    write_result_output(output, {"incident": incident})
    click.echo(f"{incident_id}: status={incident['status']}")


@evaluation.command("resolve-incident")
@click.argument("incident_id")
@click.option("--db-path", default=None, help="Learning store SQLite DB path")
@click.option("--quality-db-path", default=None, help="Quality history SQLite DB path")
@click.option("--continuous-eval-db-path", default=None, help="Continuous eval SQLite DB path")
@click.option("--prompt-registry-path", default=None, help="Prompt registry SQLite DB path")
@click.option("--actor", default=None)
@click.option("--note", default=None)
@click.option("--output", "-o", type=click.Path(path_type=Path))
def evaluation_resolve_incident(
    incident_id: str,
    db_path: str | None,
    quality_db_path: str | None,
    continuous_eval_db_path: str | None,
    prompt_registry_path: str | None,
    actor: str | None,
    note: str | None,
    output: Path | None,
) -> None:
    """Resolve an evaluation incident."""
    runner, _, _ = _build_evaluation_components(
        learning_db_path=db_path,
        quality_db_path=quality_db_path,
        continuous_eval_db_path=continuous_eval_db_path,
        prompt_registry_path=prompt_registry_path,
    )
    incident = runner.resolve_incident(incident_id, actor=actor, note=note)
    if incident is None:
        raise click.ClickException(f"Unknown incident: {incident_id}")
    write_result_output(output, {"incident": incident})
    click.echo(f"{incident_id}: status={incident['status']}")


# =============================================================================
# Improvement Group
# =============================================================================

@click.group()
def improvement() -> None:
    """Inspect true-improvement learning state for operators."""
    pass


# =============================================================================
# Transfer Group
# =============================================================================

@click.group()
def transfer() -> None:
    """Inspect and apply cross-agent learning transfers."""
    pass


__all__ = [
    "storage",
    "storage_upgrade_check",
    "storage_migrate",
    "retrieval",
    "retrieval_source_scores",
    "retrieval_reindex_triggers",
    "retrieval_profiles",
    "retrieval_ack_trigger",
    "retrieval_resolve_trigger",
    "retrieval_close_trigger",
    "retrieval_source_documents",
    "retrieval_force_reindex",
    "retrieval_export_source_scores",
    "retrieval_reindex_requests",
    "retrieval_cancel_request",
    "retrieval_resume_request",
    "retrieval_replay_request",
    "retrieval_summary",
    "tool_workflow",
    "model_routing",
    "evaluation",
    "evaluation_scorecard",
    "evaluation_regression_dataset",
    "evaluation_judge_disagreements",
    "evaluation_champion_challenger",
    "evaluation_run_regression_job",
    "evaluation_restored_jobs",
    "evaluation_trend_deltas",
    "evaluation_rollback_incidents",
    "evaluation_incident_snapshot",
    "evaluation_ack_incident",
    "evaluation_resolve_incident",
    "improvement",
    "transfer",
]



"""CachingMixin — response cache lookup and storage."""
from __future__ import annotations

from datetime import datetime, UTC

from logos_adk.types import Message, Response, content_text, normalize_content_parts


class CachingMixin:
    def _build_cache_key_data(self, provider_messages: list[Message]) -> dict:
        """Build data dict for cache key generation."""
        model_name = self._resolve_model_name()
        try:
            candidates = self._provider_candidates()
            decision = self._routing_decision_for_candidates(candidates)
            if decision is not None:
                _, _, model_name = decision.selected_model_key.partition(":")
            elif candidates:
                model_name = self._provider_model_name(candidates[0])
        except Exception:
            model_name = self._resolve_model_name()
        return {
            "model": model_name,
            "system_prompt": self._config.system_prompt,
            "messages": [
                {
                    "role": m.role,
                    "content": m.content,
                    "content_parts": [
                        {
                            "type": part.type,
                            "text": part.text,
                            "uri": part.uri,
                            "data": part.data,
                            "mime_type": part.mime_type,
                            "name": part.name,
                            "metadata": dict(part.metadata),
                        }
                        for part in normalize_content_parts(m.content_parts)
                    ],
                }
                for m in provider_messages
            ],
            "tools": self._build_tool_schemas(),
        }

    async def _cache_lookup(
        self,
        provider_messages: list[Message],
        trace_id: str,
        parent_span_id: str,
        session_id: str | None,
    ) -> Response | None:
        """Check cache for a matching response."""
        backend = self._config.cache_backend
        config = self._config.cache_config
        if backend is None or config is None:
            return None

        from logos_adk.cache.key import CacheKeyBuilder

        builder = CacheKeyBuilder()
        key_data = self._build_cache_key_data(provider_messages)

        started_at = datetime.now(UTC)
        entry = None

        if config.semantic and config.embed_fn:
            prompt = (
                content_text(
                    provider_messages[-1].content,
                    provider_messages[-1].content_parts,
                )
                if provider_messages
                else ""
            )
            vector = await config.embed_fn(prompt)
            scope = builder.scope_hash(
                model=key_data["model"],
                system_prompt=key_data["system_prompt"],
                tools=key_data["tools"],
            )
            results = await backend.search_similar(
                vector,
                config.similarity_threshold,
                limit=1,
                scope=scope,
            )
            if results:
                entry = results[0]

        if entry is None:
            key = builder.exact(**key_data)
            entry = await backend.get(key)

        hit = entry is not None
        await self._emit_span(
            name="Cache.lookup",
            trace_id=trace_id,
            parent_span_id=parent_span_id,
            started_at=started_at,
            session_id=session_id,
            attributes={
                "hit": hit,
                "backend": type(backend).__name__,
                "strategy": "semantic" if config.semantic else "exact",
            },
        )

        return entry.to_response() if entry else None

    async def _cache_store(
        self,
        provider_messages: list[Message],
        response: Response,
        trace_id: str,
        parent_span_id: str,
        session_id: str | None,
    ) -> None:
        """Store a response in the cache."""
        backend = self._config.cache_backend
        config = self._config.cache_config
        if backend is None or config is None:
            return
        if response.tool_calls:
            return

        from logos_adk.cache import CacheEntry
        from logos_adk.cache.key import CacheKeyBuilder

        builder = CacheKeyBuilder()
        key_data = self._build_cache_key_data(provider_messages)
        key = builder.exact(**key_data)

        started_at = datetime.now(UTC)
        entry = CacheEntry.from_response(response)

        if config.semantic and config.embed_fn:
            prompt = (
                content_text(
                    provider_messages[-1].content,
                    provider_messages[-1].content_parts,
                )
                if provider_messages
                else ""
            )
            entry.vector = await config.embed_fn(prompt)
            entry.scope = builder.scope_hash(
                model=key_data["model"],
                system_prompt=key_data["system_prompt"],
                tools=key_data["tools"],
            )

        await backend.put(key, entry, config.ttl)

        await self._emit_span(
            name="Cache.store",
            trace_id=trace_id,
            parent_span_id=parent_span_id,
            started_at=started_at,
            session_id=session_id,
            attributes={"key": key[:16]},
        )



"""Tests for AgentState."""
from dataclasses import dataclass, field
from datetime import UTC, datetime
import json
import sqlite3

import pytest

from logos_adk.agent import Agent
from logos_adk.config import StateConfig
from logos_adk.errors import ConfigError, StateBackendError, StateError
from logos_adk.persistence import get_sqlite_schema_version, sqlite_table_columns
from logos_adk.providers.mock import MockProvider
from logos_adk.runtime import _default_runtime_holder
from logos_adk.state import AgentState
from logos_adk.state_store import (
    InMemoryStateStore,
    ShardedStateStore,
    SQLiteStateStore,
    build_state_store,
)
from logos_adk.types import ContentPart, Message, ToolCall, ToolResult


@pytest.fixture(autouse=True)
def _reset_runtime():
    _default_runtime_holder.reset()
    yield
    _default_runtime_holder.reset()


def test_agent_state_defaults():
    state = AgentState()
    assert state.messages == []
    assert state.session_id is None
    assert state.metadata == {}


def test_agent_state_with_values():
    msg = Message(role="user", content="hello")
    state = AgentState(messages=[msg], session_id="s1", metadata={"key": "val"})
    assert len(state.messages) == 1
    assert state.session_id == "s1"
    assert state.metadata["key"] == "val"


def test_agent_state_subclass():
    @dataclass
    class ResearcherState(AgentState):
        last_results: list[dict] = field(default_factory=list)
        sources_visited: set[str] = field(default_factory=set)

    state = ResearcherState()
    assert state.messages == []
    assert state.last_results == []
    assert state.sources_visited == set()
    state.last_results.append({"url": "example.com"})
    assert len(state.last_results) == 1


def test_in_memory_state_store_round_trip_and_rollback():
    store = InMemoryStateStore()
    original = AgentState(
        messages=[Message(role="user", content="hello")],
        session_id="s1",
        metadata={"tenant": "alpha"},
    )

    saved = store.save("bot", original)
    original.messages.append(Message(role="assistant", content="mutated-after-save"))
    loaded = store.load("bot", "s1")

    assert saved.version == 1
    assert loaded is not None
    assert [message.content for message in loaded.messages] == ["hello"]

    store.save(
        "bot",
        AgentState(
            messages=[
                Message(role="user", content="hello"),
                Message(role="assistant", content="world"),
            ],
            session_id="s1",
            metadata={"tenant": "alpha"},
        ),
    )

    history = store.history("bot", "s1")
    exported = store.export_state("bot", "s1", version=1)
    rolled_back = store.rollback("bot", "s1", version=1)

    assert [item["version"] for item in history[:2]] == [2, 1]
    assert exported is not None
    assert exported["version"] == 1
    assert [message.content for message in rolled_back.messages] == ["hello"]
    assert rolled_back.version == 3

    with pytest.raises(StateError, match="No saved state version 99"):
        store.rollback("bot", "s1", version=99)


def test_sharded_state_store_import_uses_payload_session_id_for_routing():
    shard_a = InMemoryStateStore()
    shard_b = InMemoryStateStore()
    store = ShardedStateStore([shard_a, shard_b])

    payload = {
        "state": {
            "messages": [{"role": "user", "content": "restored"}],
            "session_id": "tenant-42",
            "metadata": {"source": "backup"},
            "version": 1,
            "updated_at": None,
        }
    }

    imported = store.import_state("bot", payload)

    assert imported.session_id == "tenant-42"
    assert [message.content for message in imported.messages] == ["restored"]
    assert (shard_a.export_state("bot", "tenant-42") is not None) != (
        shard_b.export_state("bot", "tenant-42") is not None
    )


def test_build_state_store_selects_expected_backends(tmp_path):
    default_store = build_state_store(None)
    persistent_store = build_state_store(StateConfig(), default_persistent=True)
    sqlite_store = build_state_store(
        StateConfig(backend="sqlite", path=str(tmp_path / "state.db"))
    )
    sharded_store = build_state_store(
        StateConfig(
            backend="sharded",
            shards=[
                StateConfig(backend="sqlite", path=str(tmp_path / "a.db")),
                StateConfig(backend="sqlite", path=str(tmp_path / "b.db")),
            ],
        )
    )

    assert isinstance(default_store, InMemoryStateStore)
    assert isinstance(persistent_store, SQLiteStateStore)
    assert isinstance(sqlite_store, SQLiteStateStore)
    assert isinstance(sharded_store, ShardedStateStore)
    assert sharded_store.describe()["shards"] == 2


def test_build_state_store_rejects_invalid_backend_configs(monkeypatch):
    monkeypatch.delenv("DATABASE_URL", raising=False)
    monkeypatch.delenv("LOGOS_ADK_DATABASE_URL", raising=False)

    with pytest.raises(ConfigError, match="requires a non-empty path"):
        build_state_store(StateConfig(backend="sqlite", path=""))

    with pytest.raises(ConfigError, match="requires a DSN"):
        build_state_store(StateConfig(backend="postgresql", dsn=None))

    with pytest.raises(ConfigError, match="requires a non-empty 'shards' list"):
        build_state_store(StateConfig(backend="sharded", shards=[]))

    with pytest.raises(ConfigError, match="Unsupported state backend"):
        build_state_store(StateConfig(backend="mystery"))


def test_sqlite_state_store_connect_wraps_backend_errors(monkeypatch):
    import sqlite3

    def _broken_connect(path: str):
        raise sqlite3.OperationalError("disk i/o error")

    monkeypatch.setattr(sqlite3, "connect", _broken_connect)

    with pytest.raises(StateBackendError, match="Failed to open SQLite state backend"):
        SQLiteStateStore("/tmp/nowhere/state.db").load("bot", "s1")


def test_sqlite_state_store_migrates_legacy_schema(tmp_path):
    db_path = tmp_path / "legacy-state.db"
    state_json = json.dumps(
        {
            "messages": [{"role": "user", "content": "hello"}],
            "session_id": "legacy-session",
            "metadata": {"source": "legacy"},
            "version": 1,
            "updated_at": "2024-01-01T00:00:00+00:00",
        }
    )

    connection = sqlite3.connect(db_path)
    try:
        connection.execute(
            """
            CREATE TABLE agent_state_versions (
                agent_name TEXT NOT NULL,
                session_key TEXT NOT NULL,
                session_id TEXT NOT NULL,
                version INTEGER NOT NULL,
                saved_at TEXT NOT NULL,
                state_json TEXT NOT NULL,
                PRIMARY KEY (agent_name, session_key, version)
            )
            """
        )
        connection.execute(
            """
            INSERT INTO agent_state_versions
            (agent_name, session_key, session_id, version, saved_at, state_json)
            VALUES (?, ?, ?, ?, ?, ?)
            """,
            (
                "bot",
                "legacy-session",
                "legacy-session",
                1,
                "2024-01-01T00:00:00+00:00",
                state_json,
            ),
        )
        connection.commit()
    finally:
        connection.close()

    store = SQLiteStateStore(str(db_path))
    loaded = store.load("bot", "legacy-session")
    exported = store.export_state("bot", "legacy-session")
    history = store.history("bot", "legacy-session")

    assert loaded is not None
    assert loaded.session_id == "legacy-session"
    assert [message.content for message in loaded.messages] == ["hello"]
    assert exported is not None
    assert exported["session_id"] == "legacy-session"
    assert exported["state"]["metadata"] == {"source": "legacy"}
    assert history[0]["version"] == 1
    assert history[0]["session_id"] == "legacy-session"

    migrated = sqlite3.connect(db_path)
    try:
        migrated.row_factory = sqlite3.Row
        columns = sqlite_table_columns(migrated, "agent_state_versions")
        assert {"namespace", "session_id", "version", "saved_at", "state_json"} <= columns
        assert "agent_name" not in columns
        assert "session_key" not in columns
        assert get_sqlite_schema_version(migrated, "state_store") == 1
    finally:
        migrated.close()


def test_sqlite_state_store_preserves_tool_payloads_and_content_parts(tmp_path):
    store = SQLiteStateStore(str(tmp_path / "state.db"))
    state = AgentState(
        messages=[
            Message(
                role="assistant",
                content="",
                content_parts=[
                    ContentPart(
                        type="image",
                        uri="https://example.com/chart.png",
                        name="chart",
                        metadata={"width": 640},
                    )
                ],
                tool_calls=[
                    ToolCall(
                        id="call-1",
                        name="search_docs",
                        arguments={"query": "incident summary", "limit": 3},
                    )
                ],
            ),
            Message(
                role="tool",
                content="tool returned",
                tool_result=ToolResult(
                    call_id="call-1",
                    content={"documents": 3},
                    content_parts=[
                        ContentPart(type="text", text="3 documents indexed"),
                        ContentPart(
                            type="file",
                            uri="file:///tmp/report.json",
                            name="report.json",
                        ),
                    ],
                    duration_ms=12.5,
                    status="completed",
                ),
            ),
        ],
        session_id="s-tooling",
        metadata={"checkpointed_at": datetime(2024, 1, 2, tzinfo=UTC)},
    )

    saved = store.save("bot", state)
    loaded = store.load("bot", "s-tooling")
    exported = store.export_state("bot", "s-tooling")

    assert saved.version == 1
    assert loaded is not None
    assert loaded.messages[0].content_parts[0].uri == "https://example.com/chart.png"
    assert loaded.messages[0].content_parts[0].metadata == {"width": 640}
    assert loaded.messages[0].tool_calls[0].arguments == {"query": "incident summary", "limit": 3}
    assert loaded.messages[1].tool_result is not None
    assert loaded.messages[1].tool_result.content == {"documents": 3}
    assert loaded.messages[1].tool_result.content_parts[1].name == "report.json"
    assert exported is not None
    assert exported["state"]["metadata"]["checkpointed_at"] == "2024-01-02T00:00:00+00:00"
    assert exported["state"]["messages"][1]["tool_result"]["content"] == {"documents": 3}


def test_sqlite_state_store_isolates_namespaces_and_clone_imports(tmp_path):
    store = SQLiteStateStore(str(tmp_path / "state.db"))

    store.save(
        "alpha",
        AgentState(
            messages=[Message(role="user", content="alpha-state")],
            session_id="shared-session",
        ),
    )
    store.save(
        "beta",
        AgentState(
            messages=[Message(role="user", content="beta-state")],
            session_id="shared-session",
        ),
    )

    alpha_export = store.export_state("alpha", "shared-session")
    beta_export = store.export_state("beta", "shared-session")
    cloned = store.import_state("alpha", alpha_export or {}, session_id="cloned-session")

    assert alpha_export is not None
    assert beta_export is not None
    assert alpha_export["state"]["messages"][0]["content"] == "alpha-state"
    assert beta_export["state"]["messages"][0]["content"] == "beta-state"
    assert cloned.session_id == "cloned-session"
    assert [message.content for message in store.load("alpha", "cloned-session").messages] == ["alpha-state"]
    assert [message.content for message in store.load("alpha", "shared-session").messages] == ["alpha-state"]
    assert [message.content for message in store.load("beta", "shared-session").messages] == ["beta-state"]
    assert [item["version"] for item in store.history("alpha", "cloned-session")] == [1]


@pytest.mark.asyncio
async def test_state_persists_across_agent_restart(tmp_path):
    store = SQLiteStateStore(str(tmp_path / "state.db"))

    first_provider = MockProvider(responses=["Hello Alex"])
    first_agent = Agent("bot").model(first_provider).state_backend(store)
    await first_agent.run("My name is Alex", session_id="s1")
    _default_runtime_holder.reset()

    second_provider = MockProvider(responses=["Your name is Alex"])
    second_agent = Agent("bot").model(second_provider).state_backend(store)
    await second_agent.run("What is my name?", session_id="s1")

    contents = [message.content for message in second_provider.call_history[0]["messages"]]
    assert "My name is Alex" in contents
    assert "Hello Alex" in contents


@pytest.mark.asyncio
async def test_state_versioning_and_rollback(tmp_path):
    store = SQLiteStateStore(str(tmp_path / "state.db"))
    provider = MockProvider(responses=["first", "second"])
    agent = Agent("bot").model(provider).state_backend(store)

    await agent.run("turn one", session_id="s1")
    await agent.run("turn two", session_id="s1")

    history = agent.list_state_versions("s1")
    assert [item["version"] for item in history[:2]] == [2, 1]

    rolled_back = agent.rollback_state("s1", version=1)
    assert rolled_back.version == 3
    assert [message.content for message in rolled_back.messages] == ["turn one", "first"]


@pytest.mark.asyncio
async def test_state_export_import_round_trip(tmp_path):
    source_store = SQLiteStateStore(str(tmp_path / "source.db"))
    target_store = SQLiteStateStore(str(tmp_path / "target.db"))

    source_agent = Agent("source").model(MockProvider(responses=["saved"])).state_backend(source_store)
    await source_agent.run("remember this", session_id="origin")
    exported = source_agent.export_state("origin")

    assert exported is not None

    target_agent = Agent("target").model(MockProvider(responses=["ok"])).state_backend(target_store)
    imported = target_agent.import_state(exported, session_id="restored")

    assert imported.session_id == "restored"
    assert [message.content for message in imported.messages] == ["remember this", "saved"]


@pytest.mark.asyncio
async def test_agent_state_helpers_expose_backend_info_and_update_hot_sessions():
    store = InMemoryStateStore()
    agent = Agent("bot").model(MockProvider(responses=["saved"])).state_backend(store)

    await agent.run("remember this", session_id="origin")
    exported = agent.export_state("origin")

    assert exported is not None
    assert agent.state_store_info()["backend"] == "in_memory"

    imported = agent.import_state(exported, session_id="clone")

    assert imported.session_id == "clone"
    assert "clone" in agent._sessions
    assert agent.list_state_versions("clone")[0]["version"] == 1


@pytest.mark.asyncio
async def test_sharded_state_store_routes_by_session(tmp_path):
    shard_a = SQLiteStateStore(str(tmp_path / "a.db"))
    shard_b = SQLiteStateStore(str(tmp_path / "b.db"))
    store = ShardedStateStore([shard_a, shard_b])

    agent = Agent("bot").model(MockProvider(responses=["alpha", "beta"])).state_backend(store)
    await agent.run("tenant-a prompt", session_id="tenant-a")
    await agent.run("tenant-b prompt", session_id="tenant-b")

    shard_a_has = shard_a.export_state("bot", "tenant-a") is not None
    shard_b_has = shard_b.export_state("bot", "tenant-a") is not None
    assert shard_a_has != shard_b_has

    shard_a_has = shard_a.export_state("bot", "tenant-b") is not None
    shard_b_has = shard_b.export_state("bot", "tenant-b") is not None
    assert shard_a_has != shard_b_has



"""Tests for adaptive memory learning."""
from __future__ import annotations

import pytest

from logos_adk.learning import MemoryLearningEngine
from logos_adk.memory import LongTermMemory, MemoryEntry, SessionMemory
from logos_adk.memory.backends import InMemoryBackend
from logos_adk.state import AgentState
from logos_adk.types import Message, Response, Usage


@pytest.mark.asyncio
async def test_memory_learning_reranks_entries_using_feedback_signals():
    backend = InMemoryBackend()
    engine = MemoryLearningEngine()

    preferred = engine.initialize_entry(MemoryEntry(content="Berlin rollout launch checklist", session_id="s1"))
    degraded = engine.initialize_entry(MemoryEntry(content="Berlin rollout launch checklist old", session_id="s1"))
    await backend.store(preferred)
    await backend.store(degraded)

    await engine.apply_feedback(backend, [preferred], helpful=True, reason="accepted_answer")
    await engine.apply_feedback(backend, [degraded], helpful=False, reason="wrong_memory")

    memory = LongTermMemory(backend=backend, retrieval="hybrid", limit=1, learning=engine)
    state = AgentState(session_id="s1")
    results = await memory.remember(
        [Message(role="user", content="launch checklist for berlin rollout")],
        state,
        session_id="s1",
    )

    assert len(results) == 1
    assert results[0].id == preferred.id
    assert state.metadata["adk_memory_entry_ids"] == [preferred.id]


@pytest.mark.asyncio
async def test_memory_learning_forgetting_policy_archives_low_value_entries():
    backend = InMemoryBackend()
    engine = MemoryLearningEngine(max_active_entries=2, archive_threshold=0.35, delete_threshold=0.1)

    low_value = engine.initialize_entry(MemoryEntry(content="stale project rumor", session_id="s1"))
    steady = engine.initialize_entry(MemoryEntry(content="current project status", session_id="s1"))
    strong = engine.initialize_entry(MemoryEntry(content="latest launch decision", session_id="s1"))
    for entry in (low_value, steady, strong):
        await backend.store(entry)

    for _ in range(3):
        await engine.apply_feedback(backend, [low_value], helpful=False, reason="bad_memory")
    await engine.apply_feedback(backend, [steady], helpful=True, reason="good_memory")
    await engine.apply_feedback(backend, [strong], helpful=True, reason="good_memory")
    await engine.enforce_forgetting_policy(backend, session_id="s1")

    remembered = await LongTermMemory(backend=backend, limit=5, learning=engine).remember(
        [Message(role="user", content="project launch")],
        AgentState(session_id="s1"),
        session_id="s1",
    )

    assert all(entry.id != low_value.id for entry in remembered)
    raw_entries = await backend.search("", session_id="s1", limit=10)
    archived = next(entry for entry in raw_entries if entry.id == low_value.id)
    assert archived.is_archived is True


@pytest.mark.asyncio
async def test_session_memory_uses_adaptive_summary_settings():
    engine = MemoryLearningEngine()
    memory = SessionMemory(strategy="summarize", max_messages=2, learning=engine)
    state = AgentState(
        session_id="s1",
        messages=[
            Message(role="user", content=f"fact {index}")
            if index % 2 == 0
            else Message(role="assistant", content=f"reply {index}")
            for index in range(16)
        ],
    )

    await memory.commit(
        state.messages,
        Response(content="reply 15", tool_calls=[], usage=Usage(1, 1), raw={}),
        state,
        session_id="s1",
    )

    assert state.metadata["adk_memory_summary_strategy"] in {"balanced", "compact"}
    assert state.messages[0].content.startswith("Session summary:\n")
    assert len(state.messages[0].content.splitlines()) <= 15


@pytest.mark.asyncio
async def test_memory_learning_adjusts_context_budget():
    backend = InMemoryBackend()
    engine = MemoryLearningEngine(max_context_limit=5)
    entries = []
    for index in range(4):
        entry = engine.initialize_entry(MemoryEntry(content=f"project signal {index}", session_id="s1"))
        await backend.store(entry)
        entries.append(entry)
    for _ in range(3):
        for entry in entries:
            await engine.apply_feedback(backend, [entry], helpful=False, reason="weak_signal")

    memory = LongTermMemory(backend=backend, retrieval="keyword", limit=2, learning=engine)
    state = AgentState(session_id="s1")
    results = await memory.remember(
        [Message(role="user", content="project signal")],
        state,
        session_id="s1",
    )

    assert state.metadata["adk_memory_context_limit"] == 4
    assert len(results) == 4



"""Evaluation loop engine for continuous learning."""
from __future__ import annotations

from dataclasses import asdict
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from logos_adk.agent import Agent
    from logos_adk.testing_framework import GoldenExample, JudgeCase, JudgeResult, LLMJudge

from .models import (
    ChampionChallengerReport,
    JudgeDisagreement,
    LearningScorecard,
    LiveChampionChallengerMetrics,
    OfflineChampionChallengerMetrics,
    ProductionRegressionDataset,
    ScoreWindow,
)
from .utils import (
    _estimate_cost_usd,
    _human_quality_score,
    _mean,
    _parse_ts,
    _quality_rows_summary,
)


class EvaluationLoopEngine:
    """Evaluation and comparison helpers built on captured production runs."""

    def __init__(self, store: Any, *, reward_scorer: Any | None = None) -> None:
        """Initialize evaluation engine.
        
        Args:
            store: LearningStore instance
            reward_scorer: Optional RewardScorer for reward-based scoring
        """
        from logos_adk.learning.dataset import DatasetBuilder
        from logos_adk.learning.improvement import RewardModel
        
        self.store = store
        self.datasets = DatasetBuilder(store)
        self.reward_scorer = reward_scorer or RewardModel()

    def resolve_reward_scorer(
        self,
        *,
        mode: str | None = None,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 500,
    ) -> Any:
        """Resolve reward scorer based on mode."""
        from logos_adk.learning.improvement import build_reward_scorer
        
        if mode is None:
            return self.reward_scorer
        return build_reward_scorer(
            mode,
            store=self.store,
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            fallback_scorer=self.reward_scorer,
        )

    def build_regression_dataset(
        self,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 100,
    ) -> ProductionRegressionDataset:
        """Build regression dataset from production runs."""
        from logos_adk.testing_framework import GoldenExample
        
        examples = self.datasets.build(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            include_unreviewed=False,
        )
        golden: list[GoldenExample] = []
        source_run_ids: list[str] = []
        for example in examples:
            expected = example.corrected_output
            if not expected:
                for feedback in example.feedback:
                    if feedback.expected_output:
                        expected = feedback.expected_output
                        break
            if not expected:
                continue
            is_bad_case = bool(example.failure_types)
            if not is_bad_case:
                is_bad_case = any(item.kind in {"thumbs_down", "edited_answer"} for item in example.feedback)
            if not is_bad_case:
                is_bad_case = any(
                    item.kind in {"escalation", "human_takeover"} and bool(item.value)
                    for item in example.outcomes
                )
            if not is_bad_case:
                continue
            golden.append(
                GoldenExample(
                    name=example.run_id,
                    prompt=example.input_text,
                    expected=expected,
                    session_id=example.session_id,
                    mode="exact",
                )
            )
            source_run_ids.append(example.run_id)
        return ProductionRegressionDataset(
            workspace_id=workspace_id,
            agent_name=agent_name,
            examples=golden,
            source_run_ids=source_run_ids,
        )

    def build_regression_suite(
        self,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 100,
    ) -> dict[str, Any]:
        """Build regression suite from production runs."""
        dataset = self.build_regression_dataset(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
        )
        return {
            "examples": [asdict(item) for item in dataset.examples],
            "source_run_ids": dataset.source_run_ids,
            "workspace_id": dataset.workspace_id,
            "agent_name": dataset.agent_name,
            "generated_at": dataset.generated_at,
        }

    def build_judge_cases_from_prod(
        self,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 100,
    ) -> list[JudgeCase]:
        """Build judge cases from production runs."""
        from logos_adk.testing_framework import JudgeCase
        
        source_examples = self.datasets.build(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            include_unreviewed=False,
        )
        cases: list[JudgeCase] = []
        for example in source_examples:
            reference = example.corrected_output
            if not reference:
                for feedback in example.feedback:
                    if feedback.expected_output:
                        reference = feedback.expected_output
                        break
            if not reference:
                continue
            criteria = "Answer should be correct, aligned with the human-corrected reference, and grounded."
            if example.failure_types:
                criteria += " Avoid failure types: " + ", ".join(sorted(example.failure_types))
            cases.append(
                JudgeCase(
                    name=example.run_id,
                    prompt=example.input_text,
                    criteria=criteria,
                    reference=reference,
                    session_id=example.session_id,
                )
            )
        return cases

    def track_judge_disagreements(
        self,
        judge_results: list[JudgeResult],
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 100,
        threshold: float = 0.35,
    ) -> list[JudgeDisagreement]:
        """Track disagreements between judge and human feedback."""
        examples = self.datasets.build(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            include_unreviewed=False,
        )
        examples_by_run = {item.run_id: item for item in examples}
        disagreements: list[JudgeDisagreement] = []
        for result in judge_results:
            example = examples_by_run.get(result.name)
            if example is None:
                continue
            human_score = _human_quality_score(example)
            delta = abs(float(result.score) - human_score)
            human_passed = human_score >= 0.6
            if delta < threshold and result.passed == human_passed:
                continue
            disagreements.append(
                JudgeDisagreement(
                    run_id=example.run_id,
                    trace_id=example.trace_id,
                    judge_score=round(float(result.score), 4),
                    human_score=human_score,
                    delta=round(delta, 4),
                    judge_passed=result.passed,
                    human_passed=human_passed,
                    prompt_version=example.prompt_version,
                    model=example.model,
                    failure_types=list(example.failure_types),
                )
            )
        return disagreements

    async def run_regression_snapshot(
        self,
        agent: Agent,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 100,
        system_prompt_override: str | None = None,
    ) -> dict[str, Any]:
        """Run regression snapshot evaluation."""
        dataset = self.build_regression_dataset(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
        )
        if system_prompt_override is None:
            from logos_adk.testing_framework import run_golden_dataset
            results = await run_golden_dataset(agent, dataset.examples)
            rows = [result.__dict__ for result in results]
        else:
            rows = []
            for example in dataset.examples:
                response = await agent.run(
                    example.prompt,
                    session_id=example.session_id,
                    system_prompt_override=system_prompt_override,
                )
                actual = response.content
                passed = example.expected in actual if example.mode == "contains" else actual == example.expected
                rows.append(
                    {
                        "name": example.name,
                        "passed": passed,
                        "actual": actual,
                        "expected": example.expected,
                        "mode": example.mode,
                    }
                )
        return {
            "kind": "regression",
            "summary": _quality_rows_summary(rows),
            "rows": rows,
            "reports": {},
        }

    async def run_judge_snapshot(
        self,
        agent: Agent,
        *,
        judge: LLMJudge,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 100,
        system_prompt_override: str | None = None,
    ) -> dict[str, Any]:
        """Run judge evaluation snapshot."""
        cases = self.build_judge_cases_from_prod(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
        )
        if system_prompt_override is None:
            from logos_adk.testing_framework import run_llm_judge_eval
            results = await run_llm_judge_eval(agent, cases, judge)
        else:
            results = []
            for case in cases:
                answer = await agent.run(
                    case.prompt,
                    session_id=case.session_id,
                    system_prompt_override=system_prompt_override,
                )
                score, rationale = await judge.evaluate(
                    prompt=case.prompt,
                    answer=answer.content,
                    criteria=case.criteria,
                    reference=case.reference,
                )
                results.append(
                    JudgeResult(
                        name=case.name,
                        score=score,
                        passed=score >= judge.pass_threshold,
                        rationale=rationale,
                        answer=answer.content,
                    )
                )
        rows = [result.__dict__ for result in results]
        disagreements = self.track_judge_disagreements(
            results,
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
        )
        summary = _quality_rows_summary(rows)
        summary["judge_disagreements"] = len(disagreements)
        return {
            "kind": "judge-eval",
            "summary": summary,
            "rows": rows,
            "reports": {},
            "disagreements": [asdict(item) for item in disagreements],
        }

    def build_scorecard(
        self,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        window_days: int = 7,
        end_at: datetime | None = None,
        limit: int = 5000,
    ) -> LearningScorecard:
        """Build learning scorecard for time window."""
        anchor = end_at or datetime.now(UTC)
        current_start = anchor - timedelta(days=window_days)
        previous_start = current_start - timedelta(days=window_days)
        runs = self.store.list_runs(workspace_id=workspace_id, agent_name=agent_name, limit=limit)
        examples = self.datasets.build(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            include_unreviewed=True,
        )
        examples_by_run = {item.run_id: item for item in examples}

        def metrics_for_window(start: datetime, end: datetime) -> tuple[dict[str, float], dict[str, int]]:
            window_runs = [
                run
                for run in runs
                if (timestamp := _parse_ts(run.created_at)) is not None and start <= timestamp < end
            ]
            window_examples = [examples_by_run[run.id] for run in window_runs if run.id in examples_by_run]
            quality_values = [
                _human_quality_score(example)
                for example in window_examples
                if example.feedback or example.outcomes
            ]
            cost_values = [_estimate_cost_usd(run.model, run.usage) for run in window_runs]
            latency_values = [
                float(run.metadata.get("latency_ms", 0.0) or 0.0)
                for run in window_runs
                if isinstance(run.metadata, dict) and run.metadata.get("latency_ms") is not None
            ]
            reviewed_examples = [example for example in window_examples if example.feedback or example.outcomes]
            bad_retrieval_count = sum(
                1 for example in reviewed_examples if "bad_retrieval" in example.failure_types
            )
            retrieval_relevance = (
                max(0.0, 1.0 - (bad_retrieval_count / len(reviewed_examples)))
                if reviewed_examples
                else 0.5
            )
            tool_runs = [run for run in window_runs if run.tool_calls]
            tool_failures = sum(
                1
                for run in tool_runs
                if "wrong_tool" in (examples_by_run.get(run.id).failure_types if run.id in examples_by_run else [])
            )
            tool_success = (
                max(0.0, 1.0 - (tool_failures / len(tool_runs)))
                if tool_runs
                else 0.5
            )
            return (
                {
                    "quality": round(_mean(quality_values, default=0.5), 4),
                    "cost": round(_mean(cost_values, default=0.0), 8),
                    "latency": round(_mean(latency_values, default=0.0), 4),
                    "retrieval_relevance": round(retrieval_relevance, 4),
                    "tool_success": round(tool_success, 4),
                },
                {
                    "runs": len(window_runs),
                    "reviewed_runs": len(reviewed_examples),
                    "tool_runs": len(tool_runs),
                },
            )

        current_metrics, current_counts = metrics_for_window(current_start, anchor)
        previous_metrics, previous_counts = metrics_for_window(previous_start, current_start)

        def window(name: str) -> ScoreWindow:
            current_value = current_metrics[name]
            previous_value = previous_metrics[name]
            return ScoreWindow(
                current=current_value,
                previous=previous_value,
                delta=round(current_value - previous_value, 4),
            )

        return LearningScorecard(
            workspace_id=workspace_id,
            agent_name=agent_name,
            window_days=window_days,
            current_window_start=current_start.isoformat(),
            current_window_end=anchor.isoformat(),
            previous_window_start=previous_start.isoformat(),
            previous_window_end=current_start.isoformat(),
            quality=window("quality"),
            cost=window("cost"),
            latency=window("latency"),
            retrieval_relevance=window("retrieval_relevance"),
            tool_success=window("tool_success"),
            counts={
                "current": current_counts,
                "previous": previous_counts,
            },
        )

    async def compare_champion_challenger(
        self,
        agent: Agent,
        *,
        challenger_prompt: str,
        champion_prompt: str | None = None,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        examples: list[GoldenExample] | None = None,
        champion_filter: dict[str, Any] | None = None,
        challenger_filter: dict[str, Any] | None = None,
        limit: int = 100,
        reward_scorer_mode: str | None = None,
        reward_training_limit: int = 500,
    ) -> ChampionChallengerReport:
        """Compare champion vs challenger configurations."""
        scorer = self.resolve_reward_scorer(
            mode=reward_scorer_mode,
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=reward_training_limit,
        )
        scorer_mode = reward_scorer_mode or "heuristic"
        champion_content = champion_prompt or getattr(getattr(agent, "_config", None), "system_prompt", "") or ""
        dataset = examples
        if dataset is None:
            dataset = self.build_regression_dataset(
                workspace_id=workspace_id,
                agent_name=agent_name,
                limit=limit,
            ).examples
        champion_offline = await self._offline_variant_metrics(
            agent,
            label="champion",
            prompt_content=champion_content,
            examples=dataset,
        )
        challenger_offline = await self._offline_variant_metrics(
            agent,
            label="challenger",
            prompt_content=challenger_prompt,
            examples=dataset,
        )
        champion_live = self._live_variant_metrics(
            label="champion",
            workspace_id=workspace_id,
            agent_name=agent_name,
            filters=champion_filter,
            limit=max(limit * 20, 500),
            reward_scorer=scorer,
        )
        challenger_live = self._live_variant_metrics(
            label="challenger",
            workspace_id=workspace_id,
            agent_name=agent_name,
            filters=challenger_filter,
            limit=max(limit * 20, 500),
            reward_scorer=scorer,
        )
        champion_score = champion_offline.pass_rate + champion_live.reward_score
        challenger_score = challenger_offline.pass_rate + challenger_live.reward_score
        recommendation = "challenger" if challenger_score >= champion_score else "champion"
        return ChampionChallengerReport(
            champion={
                "offline": asdict(champion_offline),
                "live": asdict(champion_live),
            },
            challenger={
                "offline": asdict(challenger_offline),
                "live": asdict(challenger_live),
            },
            recommendation=recommendation,
            reward_scorer_mode=scorer_mode,
        )

    async def _offline_variant_metrics(
        self,
        agent: Agent,
        *,
        label: str,
        prompt_content: str,
        examples: list[GoldenExample],
    ) -> OfflineChampionChallengerMetrics:
        """Calculate offline metrics for variant."""
        if not examples:
            return OfflineChampionChallengerMetrics(label=label, pass_rate=0.0, sample_size=0)
        results: list[float] = []
        for example in examples:
            response = await agent.run(
                example.prompt,
                session_id=example.session_id,
                system_prompt_override=prompt_content,
            )
            passed = example.expected in response.content if example.mode == "contains" else response.content == example.expected
            results.append(1.0 if passed else 0.0)
        return OfflineChampionChallengerMetrics(
            label=label,
            pass_rate=round(_mean(results, default=0.0), 4),
            sample_size=len(examples),
        )

    def _live_variant_metrics(
        self,
        *,
        label: str,
        workspace_id: str | None,
        agent_name: str | None,
        filters: dict[str, Any] | None,
        limit: int,
        reward_scorer: Any | None = None,
    ) -> LiveChampionChallengerMetrics:
        """Calculate live traffic metrics for variant."""
        runs = self.store.list_runs(workspace_id=workspace_id, agent_name=agent_name, limit=limit)
        matched = [run for run in runs if self._matches_filters(run, filters or {})]
        examples = {
            item.run_id: item
            for item in self.datasets.build(
                workspace_id=workspace_id,
                agent_name=agent_name,
                limit=limit,
                include_unreviewed=True,
            )
        }
        quality_values: list[float] = []
        cost_values: list[float] = []
        latency_values: list[float] = []
        conversion_values: list[float] = []
        reward_values: list[float] = []
        scorer = reward_scorer or self.reward_scorer
        for run in matched:
            example = examples.get(run.id)
            quality = _human_quality_score(example) if example is not None else 0.5
            cost = _estimate_cost_usd(run.model, run.usage)
            latency = 0.0
            if isinstance(run.metadata, dict):
                latency = float(run.metadata.get("latency_ms", 0.0) or 0.0)
            conversions = 0.0
            if example is not None:
                for outcome in example.outcomes:
                    if outcome.kind == "conversion" and bool(outcome.value):
                        conversions = 1.0
                        break
            if example is not None:
                reward = scorer.score_example(example).score
            else:
                reward = scorer.score_run(
                    run,
                    feedback=self.store.list_feedback(run_id=run.id),
                    outcomes=self.store.list_outcomes(run_id=run.id),
                ).score
            quality_values.append(quality)
            cost_values.append(cost)
            latency_values.append(latency)
            conversion_values.append(conversions)
            reward_values.append(reward)
        return LiveChampionChallengerMetrics(
            label=label,
            runs=len(matched),
            quality_score=round(_mean(quality_values, default=0.0), 4),
            avg_cost_usd=round(_mean(cost_values, default=0.0), 8),
            avg_latency_ms=round(_mean(latency_values, default=0.0), 4),
            conversion_rate=round(_mean(conversion_values, default=0.0), 4),
            reward_score=round(_mean(reward_values, default=0.0), 4),
        )

    def _matches_filters(self, run: Any, filters: dict[str, Any]) -> bool:
        """Check if run matches filters."""
        metadata = getattr(run, "metadata", {}) or {}
        for key, expected in filters.items():
            if key in {"prompt_version", "model", "config_fingerprint"}:
                if getattr(run, key, None) != expected:
                    return False
            elif metadata.get(key) != expected:
                return False
        return True


__all__ = ["EvaluationLoopEngine"]



from __future__ import annotations

import os

from logos_adk.dynamic_orchestration import DepartmentSupervisor
from logos_adk.runnable_config_loader import (
    describe_runnable_from_config,
    detect_runnable_config_kind,
    load_runnable_from_config,
)
from logos_adk.team import Team


FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")


def test_detect_runnable_config_kind_for_department_supervisor():
    path = os.path.join(FIXTURES, "department_supervisor.yaml")

    assert detect_runnable_config_kind(path) == "department_supervisor"


def test_load_runnable_from_config_supports_team():
    path = os.path.join(FIXTURES, "team_event_bus.yaml")
    runnable = load_runnable_from_config(path)

    assert isinstance(runnable, Team)
    assert runnable.name == "seo_team"


def test_load_runnable_from_config_supports_department_supervisor():
    path = os.path.join(FIXTURES, "department_supervisor.yaml")
    runnable = load_runnable_from_config(path)

    assert isinstance(runnable, DepartmentSupervisor)
    assert runnable.name == "ceo"
    assert describe_runnable_from_config(path) == {
        "kind": "department_supervisor",
        "name": "ceo",
    }



"""Dynamic department-level orchestration across teams, crews, and swarms."""
from __future__ import annotations

import importlib
import json
import re
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Literal
from uuid import uuid4

from pydantic import BaseModel, Field

from logos_adk.crew import CrewProcess
from logos_adk.dynamic_orchestration_config import DepartmentSupervisorConfig
from logos_adk.errors import ConfigError
from logos_adk.event_bus import InMemoryEventBus
from logos_adk.runnable import Runnable
from logos_adk.subscription_registry import subscription_handler_registry
from logos_adk.swarm import Swarm
from logos_adk.team import Team
from logos_adk.types import Response, StreamChunk, StreamDone, StructuredResponse, TextChunk
from logos_adk.ui_events import make_ui_event


DepartmentKind = Literal["team", "crew", "swarm", "runnable"]


class DepartmentRoutingDecisionSchema(BaseModel):
    """Structured supervisor decision for department selection."""

    department_name: str = Field(min_length=1)
    delegated_prompt: str | None = None
    reason: str | None = None


@dataclass
class DepartmentSpec:
    """Descriptor for one executable department under the supervisor."""

    name: str
    target: Runnable | Swarm
    description: str = ""
    kind: DepartmentKind = "runnable"
    capability_tags: list[str] = field(default_factory=list)
    metadata: dict[str, Any] = field(default_factory=dict)
    request_topic: str | None = None
    response_topic: str | None = None


class DepartmentSupervisor(Runnable):
    """LLM-guided supervisor that routes goals to the most suitable department."""

    def __init__(
        self,
        name: str,
        *,
        router: Any | None = None,
        departments: list[DepartmentSpec] | None = None,
        fallback_department: str | None = None,
    ) -> None:
        self._name = name
        self._router = router
        self._fallback_department = fallback_department
        self._departments: dict[str, DepartmentSpec] = {}
        for department in departments or []:
            self.add_department_spec(department)

    @property
    def name(self) -> str:
        return self._name

    def department(
        self,
        name: str,
        target: Runnable | Swarm,
        *,
        description: str = "",
        kind: DepartmentKind | None = None,
        capability_tags: list[str] | None = None,
        metadata: dict[str, Any] | None = None,
        request_topic: str | None = None,
        response_topic: str | None = None,
    ) -> DepartmentSupervisor:
        return self.add_department_spec(
            DepartmentSpec(
                name=name,
                target=target,
                description=description,
                kind=kind or _infer_department_kind(target),
                capability_tags=list(capability_tags or []),
                metadata=dict(metadata or {}),
                request_topic=request_topic,
                response_topic=response_topic,
            )
        )

    def add_department_spec(self, department: DepartmentSpec) -> DepartmentSupervisor:
        if department.name in self._departments:
            raise ValueError(f"Department '{department.name}' is already registered")
        self._departments[department.name] = department
        if self._fallback_department is None:
            self._fallback_department = department.name
        return self

    def list_departments(self) -> list[DepartmentSpec]:
        return list(self._departments.values())

    @classmethod
    def from_definition(cls, config: DepartmentSupervisorConfig) -> DepartmentSupervisor:
        """Build a DepartmentSupervisor from a typed config."""
        from logos_adk.agent import Agent
        from logos_adk.config_loader import build_agent_from_definition

        router = (
            build_agent_from_definition(config.router, default_cls=Agent)
            if config.router is not None
            else None
        )
        supervisor = cls(
            config.name,
            router=router,
            fallback_department=config.fallback_department,
        )
        for department in config.departments:
            if department.kind == "team":
                if department.team is None:
                    raise ConfigError(f"Department '{department.name}' requires a team config")
                target = Team.from_definition(department.team)
            elif department.kind == "crew":
                if department.team is None:
                    raise ConfigError(f"Department '{department.name}' requires a team config")
                team = Team.from_definition(department.team)
                target = CrewProcess(
                    department.name,
                    team=team,
                    tasks=department.tasks,
                    process=department.process,
                    manager_role=department.manager_role,
                    max_parallel_tasks=department.max_parallel_tasks,
                )
            elif department.kind == "swarm":
                target = Swarm(
                    department.name,
                    event_bus=Team._build_event_bus_from_config(department.event_bus) or InMemoryEventBus(),
                    retry_delay_seconds=department.retry_delay_seconds,
                    lease_seconds=department.lease_seconds,
                )
                for subscription in department.subscriptions:
                    target.subscribe(
                        subscription.topic,
                        _resolve_swarm_handler(subscription.handler),
                        name=subscription.name,
                        publish_to=subscription.publish_to,
                    )
            else:
                raise ConfigError(f"Unsupported department kind '{department.kind}'")
            supervisor.department(
                department.name,
                target,
                description=department.description,
                kind=department.kind,
                capability_tags=department.capability_tags,
                metadata=department.metadata,
                request_topic=department.request_topic,
                response_topic=department.response_topic,
            )
        return supervisor

    @classmethod
    def from_config(cls, path: str) -> DepartmentSupervisor:
        """Load a DepartmentSupervisor from a YAML config file."""
        from logos_adk.dynamic_orchestration_config_parser import load_department_supervisor_config

        return cls.from_definition(load_department_supervisor_config(path))

    async def run(
        self,
        prompt: str,
        *,
        session_id: str | None = None,
        **kwargs: Any,
    ) -> Response:
        department, decision = await self._select_department(prompt, session_id=session_id)
        delegated_prompt = (decision.delegated_prompt or prompt).strip() or prompt
        response = await self._run_department(
            department,
            delegated_prompt,
            session_id=session_id,
            **kwargs,
        )
        raw = dict(response.raw or {})
        raw["dynamic_orchestration"] = {
            "supervisor": self._name,
            "department_name": department.name,
            "department_kind": department.kind,
            "delegated_prompt": delegated_prompt,
            "reason": decision.reason,
        }
        return Response(
            content=response.content,
            content_parts=response.content_parts,
            tool_calls=response.tool_calls,
            usage=response.usage,
            raw=raw,
            reasoning=response.reasoning,
            status=response.status,
        )

    async def stream(self, prompt: str, **kwargs: Any) -> AsyncIterator[StreamChunk]:
        response = await self.run(prompt, session_id=kwargs.pop("session_id", None), **kwargs)
        if response.content:
            yield TextChunk(text=response.content, reasoning=response.reasoning)
        yield StreamDone(response=response)

    async def stream_events(
        self,
        prompt: str,
        *,
        session_id: str | None = None,
        correlation_id: str | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[dict[str, Any]]:
        run_id = uuid4().hex
        sequence = 0
        trace_id = kwargs.get("_trace_id")
        effective_correlation_id = correlation_id or session_id or run_id
        yield make_ui_event(
            event="graph_started",
            event_type="supervisor",
            graph_name=self._name,
            graph_run_id=run_id,
            sequence=sequence,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            trace_id=trace_id,
            node_name=self._name,
            status="running",
        )
        sequence += 1
        yield make_ui_event(
            event="node_started",
            event_type="routing",
            graph_name=self._name,
            graph_run_id=run_id,
            sequence=sequence,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            trace_id=trace_id,
            node_name="router",
            status="running",
            metadata={
                "user_facing": {
                    "kind": "step",
                    "title": "Выбираем маршрут",
                    "detail": "Определяем подходящий отдел для выполнения задачи",
                    "step_key": "route:select",
                    "status": "running",
                    "visible": True,
                }
            },
        )
        sequence += 1
        department, decision = await self._select_department(prompt, session_id=session_id)
        delegated_prompt = (decision.delegated_prompt or prompt).strip() or prompt
        yield make_ui_event(
            event="node_completed",
            event_type="routing",
            graph_name=self._name,
            graph_run_id=run_id,
            sequence=sequence,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            trace_id=trace_id,
            node_name="router",
            status="completed",
            output={
                "department_name": department.name,
                "department_kind": department.kind,
                "delegated_prompt": delegated_prompt,
                "reason": decision.reason,
            },
            metadata={
                "user_facing": {
                    "kind": "step",
                    "title": "Маршрут выбран",
                    "detail": f"Передаём задачу в {department.name}",
                    "step_key": "route:select",
                    "status": "completed",
                    "visible": True,
                }
            },
        )
        sequence += 1
        target = department.target
        if isinstance(target, Swarm):
            final_event: dict[str, Any] | None = None
            async for event in target.dispatch_events(
                delegated_prompt,
                request_topic=department.request_topic or department.name,
                response_topic=department.response_topic,
                session_id=session_id,
                correlation_id=effective_correlation_id,
                trace_id=trace_id,
            ):
                payload = _decorate_forwarded_event(event, department_name=department.name)
                if payload is None:
                    continue
                if payload.get("event") == "graph_completed":
                    final_event = payload
                    continue
                yield payload
            yield _build_supervisor_terminal_event(
                supervisor_name=self._name,
                run_id=run_id,
                sequence=sequence,
                session_id=session_id,
                correlation_id=effective_correlation_id,
                trace_id=(final_event or {}).get("trace_id") or trace_id,
                department=department,
                decision=decision,
                delegated_prompt=delegated_prompt,
                final_event=final_event,
            )
            return
        if hasattr(target, "stream_ui_events"):
            final_event = None
            async for event in target.stream_ui_events(
                delegated_prompt,
                session_id=session_id,
                correlation_id=effective_correlation_id,
                **kwargs,
            ):
                payload = _decorate_forwarded_event(event, department_name=department.name)
                if payload is None:
                    continue
                if payload.get("event") == "graph_completed":
                    final_event = payload
                    continue
                yield payload
            yield _build_supervisor_terminal_event(
                supervisor_name=self._name,
                run_id=run_id,
                sequence=sequence,
                session_id=session_id,
                correlation_id=effective_correlation_id,
                trace_id=(final_event or {}).get("trace_id") or trace_id,
                department=department,
                decision=decision,
                delegated_prompt=delegated_prompt,
                final_event=final_event,
            )
            return
        if hasattr(target, "stream_events"):
            final_event = None
            async for event in target.stream_events(
                delegated_prompt,
                session_id=session_id,
                correlation_id=effective_correlation_id,
                **kwargs,
            ):
                payload = _decorate_forwarded_event(event, department_name=department.name)
                if payload is None:
                    continue
                if payload.get("event") == "graph_completed":
                    final_event = payload
                    continue
                yield payload
            yield _build_supervisor_terminal_event(
                supervisor_name=self._name,
                run_id=run_id,
                sequence=sequence,
                session_id=session_id,
                correlation_id=effective_correlation_id,
                trace_id=(final_event or {}).get("trace_id") or trace_id,
                department=department,
                decision=decision,
                delegated_prompt=delegated_prompt,
                final_event=final_event,
            )
            return
        response = await self._run_department(
            department,
            delegated_prompt,
            session_id=session_id,
            **kwargs,
        )
        yield make_ui_event(
            event="graph_completed",
            event_type="supervisor",
            graph_name=self._name,
            graph_run_id=run_id,
            sequence=sequence,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            trace_id=trace_id,
            node_name=department.name,
            status=response.status,
            output=response,
        )

    async def _select_department(
        self,
        prompt: str,
        *,
        session_id: str | None,
    ) -> tuple[DepartmentSpec, DepartmentRoutingDecisionSchema]:
        if not self._departments:
            raise ValueError("DepartmentSupervisor requires at least one department")
        if len(self._departments) == 1:
            only = next(iter(self._departments.values()))
            return only, DepartmentRoutingDecisionSchema(
                department_name=only.name,
                delegated_prompt=prompt,
                reason="Only one department is registered.",
            )
        routed = await self._route_with_router(prompt, session_id=session_id)
        if routed is not None:
            department = self._departments.get(routed.department_name)
            if department is not None:
                return department, routed
        fallback = self._heuristic_department(prompt)
        return fallback, DepartmentRoutingDecisionSchema(
            department_name=fallback.name,
            delegated_prompt=prompt,
            reason="Heuristic routing fallback.",
        )

    async def _route_with_router(
        self,
        prompt: str,
        *,
        session_id: str | None,
    ) -> DepartmentRoutingDecisionSchema | None:
        if self._router is None or not hasattr(self._router, "run"):
            return None
        routing_prompt = self._build_routing_prompt(prompt)
        try:
            response = await self._router.run(
                routing_prompt,
                session_id=session_id,
                output_schema=DepartmentRoutingDecisionSchema,
            )
        except TypeError:
            response = await self._router.run(routing_prompt, session_id=session_id)
        if isinstance(response, StructuredResponse) and response.parsed is not None:
            return response.parsed
        try:
            return DepartmentRoutingDecisionSchema.model_validate_json(response.content)
        except Exception:
            return None

    def _build_routing_prompt(self, prompt: str) -> str:
        lines = [
            "You are routing a business goal to the best department.",
            "Return strict JSON with keys: department_name, delegated_prompt, reason.",
            "Choose exactly one department from the list below.",
            "Departments:",
        ]
        for department in self._departments.values():
            tags = ", ".join(department.capability_tags) if department.capability_tags else "-"
            lines.append(
                f"- {department.name} [{department.kind}] tags={tags} description={department.description or '-'}"
            )
        lines.extend(
            [
                "",
                "Business goal:",
                prompt,
            ]
        )
        return "\n".join(lines)

    def _heuristic_department(self, prompt: str) -> DepartmentSpec:
        prompt_tokens = _score_tokens(prompt)
        scored = [
            (
                _department_score(prompt_tokens, department, prompt),
                index,
                department,
            )
            for index, department in enumerate(self._departments.values())
        ]
        scored.sort(key=lambda item: (item[0], -item[1]), reverse=True)
        top_score = scored[0][0]
        if top_score <= 0 and self._fallback_department is not None:
            return self._departments[self._fallback_department]
        return scored[0][2]

    async def _run_department(
        self,
        department: DepartmentSpec,
        prompt: str,
        *,
        session_id: str | None,
        **kwargs: Any,
    ) -> Response:
        if isinstance(department.target, Swarm):
            request_topic = department.request_topic or department.name
            return await department.target.dispatch(
                prompt,
                request_topic=request_topic,
                response_topic=department.response_topic,
                session_id=session_id,
                campaign_id=kwargs.pop("campaign_id", None),
                run_id=kwargs.pop("run_id", None),
            )
        return await department.target.run(prompt, session_id=session_id, **kwargs)


def _infer_department_kind(target: Runnable | Swarm) -> DepartmentKind:
    if isinstance(target, Swarm):
        return "swarm"
    target_name = type(target).__name__
    if target_name == "Team":
        return "team"
    if target_name == "CrewProcess":
        return "crew"
    return "runnable"


def _decorate_forwarded_event(
    event: Any,
    *,
    department_name: str | None = None,
    team_name: str | None = None,
) -> dict[str, Any] | None:
    payload = _normalize_routed_event(event)
    payload.setdefault("metadata", {})
    if department_name is not None:
        payload["metadata"]["department_name"] = department_name
    if team_name is not None:
        payload["metadata"]["team_name"] = team_name
    if payload.get("event") == "graph_started":
        return None
    return payload


def _build_supervisor_terminal_event(
    *,
    supervisor_name: str,
    run_id: str,
    sequence: int,
    session_id: str | None,
    correlation_id: str | None,
    trace_id: str | None,
    department: DepartmentSpec,
    decision: DepartmentRoutingDecisionSchema,
    delegated_prompt: str,
    final_event: dict[str, Any] | None,
) -> dict[str, Any]:
    metadata = dict((final_event or {}).get("metadata") or {})
    metadata["department_name"] = department.name
    metadata["department_kind"] = department.kind
    metadata["delegated_prompt"] = delegated_prompt
    metadata["reason"] = decision.reason
    if final_event is None:
        metadata["user_facing"] = {
            "kind": "error",
            "title": "Не удалось завершить задачу",
            "detail": f"Отдел '{department.name}' завершил поток без итогового terminal event.",
            "step_key": "run:department_missing_terminal",
            "status": "failed",
            "visible": True,
        }
        output = {
            "error": "department_missing_terminal",
            "department_name": department.name,
            "department_kind": department.kind,
            "reason": decision.reason,
        }
        status = "failed"
    else:
        output = final_event.get("output")
        status = final_event.get("status") or "completed"
    return make_ui_event(
        event="graph_completed",
        event_type="supervisor",
        graph_name=supervisor_name,
        graph_run_id=run_id,
        sequence=sequence,
        session_id=session_id,
        correlation_id=correlation_id,
        trace_id=trace_id,
        node_name=department.name,
        status=status,
        output=output,
        metadata=metadata,
    )


def _normalize_routed_event(event: Any) -> dict[str, Any]:
    if isinstance(event, dict):
        return dict(event)
    from logos_adk.workflow.graph import GraphNodeEvent, _serialize_graph_ui_event

    if isinstance(event, GraphNodeEvent):
        return _serialize_graph_ui_event(event)
    raise TypeError(f"Unsupported routed event type: {type(event)!r}")


def _score_tokens(value: str) -> set[str]:
    return {
        token
        for token in re.split(r"[^a-zA-Z0-9_]+", value.lower())
        if len(token) >= 3
    }


def _department_score(prompt_tokens: set[str], department: DepartmentSpec, prompt_text: str) -> int:
    haystacks = [
        department.name,
        department.description,
        " ".join(department.capability_tags),
        json.dumps(department.metadata, ensure_ascii=False, sort_keys=True),
    ]
    haystack_tokens: set[str] = set()
    for haystack in haystacks:
        haystack_tokens.update(_score_tokens(haystack))
    score = len(prompt_tokens & haystack_tokens)
    if department.name.lower() in prompt_text.lower():
        score += 2
    return score


def _resolve_swarm_handler(handler: str) -> Any:
    registered = subscription_handler_registry.get(handler)
    if registered is not None:
        return registered
    imported = _load_handler_from_path(handler)
    if imported is not None:
        return imported
    raise ConfigError(
        f"Unknown swarm handler '{handler}'. Register it in subscription_handler_registry "
        "or provide an import path like 'package.module:handler'."
    )


def _load_handler_from_path(path: str) -> Any | None:
    if ":" in path:
        module_name, attr_name = path.split(":", 1)
    elif "." in path:
        module_name, attr_name = path.rsplit(".", 1)
    else:
        return None
    if not module_name or not attr_name:
        return None
    try:
        module = importlib.import_module(module_name)
    except Exception:
        return None
    candidate = getattr(module, attr_name, None)
    return candidate if callable(candidate) else None


__all__ = [
    "DepartmentKind",
    "DepartmentRoutingDecisionSchema",
    "DepartmentSpec",
    "DepartmentSupervisor",
]



"""Closed-loop self-learning orchestration built on existing learning primitives."""
from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Any
import json

from logos_adk.agent import Agent

from .dataset import DatasetBuilder
from .evaluation import ContinuousEvalRunner, EvaluationLoopEngine
from .improvement import FailurePattern, PatternMiner, ReflectionMemory, RewardModel, RewardScorer, build_reward_scorer
from .optimizer import PromptCandidateScore, PromptOptimizer
from .prompt_evolution import PromptCandidate, PromptCandidateGenerator, PromptGenerationReport
from .prompt_registry import PromptRegistry
from .store import LearningStore


@dataclass
class SelfLearningLoopPolicy:
    """Thresholds and rollout policy for one self-learning run."""

    min_sample_size: int = 3
    min_improvement: float = 0.05
    max_candidates: int = 5
    feedback_limit: int = 100
    rollout_percentage: int = 20
    environment: str = "prod"
    allow_auto_promote: bool = False
    max_latency_increase_ratio: float = 1.15
    max_cost_increase_ratio: float = 1.15
    auto_rollback_on_regression: bool = True
    rollback_min_pass_rate: float = 0.95
    rollback_min_total_cases: int = 1
    disable_experiment_after: int = 2
    continuous_eval_interval_seconds: float = 3600.0
    continuous_eval_limit: int = 100
    continuous_eval_job_name: str | None = None
    reward_scorer_mode: str = "heuristic"
    reward_training_limit: int = 500


@dataclass
class SelfLearningLoopAction:
    """One deploy/experiment/noop decision from the loop."""

    kind: str
    status: str
    reason: str
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class SelfLearningLoopRun:
    """Structured result for a self-learning iteration."""

    prompt_name: str
    workspace_id: str | None
    task_type: str | None
    user_segment: str | None
    dry_run: bool
    reward_scorer_mode: str
    baseline_version: int
    baseline_score: float
    winner_score: float
    improvement: float
    historical_reward_score: float
    failure_count: int
    pattern_count: int
    candidate_count: int
    generation_report: dict[str, Any] | None = None
    scores: list[PromptCandidateScore] = field(default_factory=list)
    patterns: list[FailurePattern] = field(default_factory=list)
    action: SelfLearningLoopAction | None = None
    created_version: dict[str, Any] | None = None
    experiment: dict[str, Any] | None = None
    deployment: dict[str, Any] | None = None
    audit_trail: list[str] = field(default_factory=list)


class SelfLearningLoop:
    """Run a minimal self-improvement cycle for prompt evolution."""

    def __init__(
        self,
        store: LearningStore,
        registry: PromptRegistry,
        *,
        prompt_optimizer: PromptOptimizer | None = None,
        evaluation_loop: EvaluationLoopEngine | None = None,
        continuous_eval_runner: ContinuousEvalRunner | None = None,
        pattern_miner: PatternMiner | None = None,
        reflection_memory: ReflectionMemory | None = None,
        candidate_generator: PromptCandidateGenerator | None = None,
        reward_scorer: RewardScorer | None = None,
    ) -> None:
        self.store = store
        self.registry = registry
        self.prompt_optimizer = prompt_optimizer or PromptOptimizer(store, registry)
        self.reward_scorer = reward_scorer or RewardModel()
        self.evaluation_loop = evaluation_loop or EvaluationLoopEngine(store, reward_scorer=self.reward_scorer)
        self.continuous_eval_runner = continuous_eval_runner
        self.pattern_miner = pattern_miner or PatternMiner(store)
        self.reflection_memory = reflection_memory or ReflectionMemory()
        self.candidate_generator = candidate_generator or PromptCandidateGenerator(
            registry,
            store=store,
            reflection_memory=self.reflection_memory,
        )
        self.datasets = DatasetBuilder(store)

    def collect_failures(
        self,
        *,
        prompt_name: str,
        workspace_id: str | None = None,
        limit: int = 100,
    ) -> list[Any]:
        examples = self.datasets.build(
            workspace_id=workspace_id,
            agent_name=prompt_name,
            limit=limit,
            include_unreviewed=False,
        )
        return [item for item in examples if item.failure_types or item.corrected_output]

    def mine_patterns(
        self,
        *,
        prompt_name: str,
        workspace_id: str | None = None,
        limit: int = 100,
    ) -> list[FailurePattern]:
        return self.pattern_miner.mine_patterns(
            workspace_id=workspace_id,
            agent_name=prompt_name,
            limit=limit,
            min_occurrences=1,
        )

    def generate_candidates(
        self,
        *,
        prompt_name: str,
        workspace_id: str | None = None,
        task_type: str | None = None,
        user_segment: str | None = None,
        environment: str = "prod",
        patterns: list[FailurePattern] | None = None,
        max_candidates: int = 5,
    ) -> list[PromptCandidate]:
        return self.candidate_generator.generate(
            prompt_name=prompt_name,
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            environment=environment,
            patterns=patterns,
            max_candidates=max_candidates,
        )

    async def evaluate_candidates(
        self,
        agent: Agent,
        *,
        prompt_name: str,
        candidates: list[str],
        workspace_id: str | None = None,
        feedback_limit: int = 100,
    ) -> list[PromptCandidateScore]:
        regression = self.evaluation_loop.build_regression_dataset(
            workspace_id=workspace_id,
            agent_name=prompt_name,
            limit=feedback_limit,
        )
        return await self.prompt_optimizer.evaluate_candidates(
            agent,
            prompt_name=prompt_name,
            candidates=candidates,
            golden_examples=regression.examples,
            workspace_id=workspace_id,
            feedback_limit=feedback_limit,
        )

    async def run_once(
        self,
        agent: Agent,
        *,
        prompt_name: str,
        workspace_id: str | None = None,
        task_type: str | None = None,
        user_segment: str | None = None,
        policy: SelfLearningLoopPolicy | None = None,
        dry_run: bool = False,
    ) -> SelfLearningLoopRun:
        policy = policy or SelfLearningLoopPolicy()
        resolution = self.registry.resolve(
            prompt_name,
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            environment=policy.environment,
        )
        if resolution is None:
            versions = self.registry.list_versions(prompt_name)
            if not versions:
                raise ValueError(f"Prompt not found: {prompt_name}")
            baseline_version = versions[0].version
            baseline_content = versions[0].content
        else:
            baseline_version = resolution.version
            baseline_content = resolution.content

        failures = self.collect_failures(
            prompt_name=prompt_name,
            workspace_id=workspace_id,
            limit=policy.feedback_limit,
        )
        patterns = self.mine_patterns(
            prompt_name=prompt_name,
            workspace_id=workspace_id,
            limit=policy.feedback_limit,
        )
        generation_report = self.candidate_generator.generate_report(
            prompt_name=prompt_name,
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            environment=policy.environment,
            patterns=patterns,
            max_candidates=policy.max_candidates,
        )
        candidates = generation_report.candidates
        run = SelfLearningLoopRun(
            prompt_name=prompt_name,
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            dry_run=dry_run,
            reward_scorer_mode=policy.reward_scorer_mode,
            baseline_version=baseline_version,
            baseline_score=0.0,
            winner_score=0.0,
            improvement=0.0,
            historical_reward_score=0.0,
            failure_count=len(failures),
            pattern_count=len(patterns),
            candidate_count=len(candidates),
            generation_report=self._serialize_generation_report(generation_report),
            patterns=patterns,
        )
        reward_scorer = self._resolve_reward_scorer(
            mode=policy.reward_scorer_mode,
            workspace_id=workspace_id,
            prompt_name=prompt_name,
            limit=policy.reward_training_limit,
        )
        run.historical_reward_score = self._historical_reward_score(failures, reward_scorer)
        run.audit_trail.append(f"reward_scorer_mode={policy.reward_scorer_mode}")
        run.audit_trail.append(f"historical_reward={run.historical_reward_score}")

        if len(failures) < policy.min_sample_size:
            run.action = SelfLearningLoopAction(
                kind="noop",
                status="skipped",
                reason="insufficient_failures",
                metadata={"failure_count": len(failures), "min_sample_size": policy.min_sample_size},
            )
            run.audit_trail.append("insufficient_failures")
            return run

        if not candidates:
            run.action = SelfLearningLoopAction(
                kind="noop",
                status="skipped",
                reason="no_candidates_generated",
            )
            run.audit_trail.append("no_candidates_generated")
            return run

        persisted_candidates = self.candidate_generator.persist_candidates(generation_report)
        run.audit_trail.append(f"generated_candidates={len(persisted_candidates)}")

        candidate_map = {candidate.content: candidate for candidate in candidates}
        scores = await self.evaluate_candidates(
            agent,
            prompt_name=prompt_name,
            candidates=[baseline_content, *candidate_map.keys()],
            workspace_id=workspace_id,
            feedback_limit=policy.feedback_limit,
        )
        run.scores = scores
        baseline_score = next((item for item in scores if item.content == baseline_content), None)
        winner_score = scores[0]
        run.baseline_score = baseline_score.combined_score if baseline_score is not None else 0.0
        run.winner_score = winner_score.combined_score
        run.improvement = round(run.winner_score - run.baseline_score, 4)

        if winner_score.content == baseline_content:
            run.action = SelfLearningLoopAction(
                kind="noop",
                status="skipped",
                reason="baseline_still_best",
            )
            run.audit_trail.append("baseline_still_best")
            return run
        if run.improvement < policy.min_improvement:
            run.action = SelfLearningLoopAction(
                kind="noop",
                status="skipped",
                reason="insufficient_lift",
                metadata={"improvement": run.improvement, "min_improvement": policy.min_improvement},
            )
            run.audit_trail.append("insufficient_lift")
            return run

        selected = candidate_map[winner_score.content]
        baseline_metrics = self._live_metrics(
            prompt_name=prompt_name,
            prompt_version=f"v{baseline_version}",
            workspace_id=workspace_id,
        )
        candidate_metrics = self._project_candidate_metrics(
            baseline_metrics=baseline_metrics,
            baseline_content=baseline_content,
            candidate=selected,
            patterns=patterns,
        )
        guard = self.prompt_optimizer.guard.evaluate(
            baseline_metrics=baseline_metrics,
            candidate_metrics=candidate_metrics,
            max_latency_increase_ratio=policy.max_latency_increase_ratio,
            max_cost_increase_ratio=policy.max_cost_increase_ratio,
        )
        run.audit_trail.append(f"winner_strategy={selected.strategy}")
        run.audit_trail.append(f"lift={run.improvement}")
        if not guard.allowed:
            run.audit_trail.append(f"guard_blocked={','.join(guard.reasons)}")
        if dry_run:
            run.action = SelfLearningLoopAction(
                kind="promote" if policy.allow_auto_promote and guard.allowed else "experiment",
                status="dry_run",
                reason="dry_run",
                metadata={"guard_allowed": guard.allowed, "guard_reasons": guard.reasons},
            )
            run.audit_trail.append("dry_run")
            return run

        provenance = {
            "source": "self_learning_loop",
            "strategy": selected.strategy,
            "rationale": selected.rationale,
            "improvement": run.improvement,
            "reward_scorer_mode": policy.reward_scorer_mode,
            "historical_reward_score": run.historical_reward_score,
            "guard_allowed": guard.allowed,
            "patterns": [pattern.signature for pattern in patterns[:5]],
            "generated_candidates": self._serialize_generation_report(generation_report)["candidates"],
        }
        rollback_policy = {
            "prompt_name": prompt_name,
            "workspace_id": workspace_id,
            "task_type": task_type,
            "user_segment": user_segment,
            "environment": policy.environment,
            "auto_rollback_on_regression": policy.auto_rollback_on_regression,
            "min_pass_rate": policy.rollback_min_pass_rate,
            "min_total_cases": policy.rollback_min_total_cases,
            "disable_experiment_after": policy.disable_experiment_after,
            "reward_scorer_mode": policy.reward_scorer_mode,
        }
        created = self.registry.create_version(
            prompt_name,
            content=selected.content,
            tags=["self-learning", selected.strategy],
            notes=json.dumps({**provenance, "rollback_policy": rollback_policy}, sort_keys=True),
            created_by="self_learning_loop",
        )
        run.created_version = asdict(created)

        if policy.allow_auto_promote and guard.allowed:
            deployment = self.registry.promote(
                prompt_name,
                version=created.version,
                environment=policy.environment,
                workspace_id=workspace_id,
                task_type=task_type,
                user_segment=user_segment,
                metadata={
                    "source": "self_learning_loop",
                    "baseline_version": baseline_version,
                    "improvement": run.improvement,
                    "rollback_policy": rollback_policy,
                },
            )
            run.deployment = asdict(deployment)
            scheduled_job = self._ensure_continuous_eval_job(
                agent,
                prompt_name=prompt_name,
                workspace_id=workspace_id,
                task_type=task_type,
                user_segment=user_segment,
                policy=policy,
                rollback_policy=rollback_policy,
                target_version=created.version,
            )
            run.action = SelfLearningLoopAction(
                kind="promote",
                status="executed",
                reason="auto_promoted",
                metadata={"version": created.version, "continuous_eval_job": scheduled_job},
            )
            run.audit_trail.append(f"promoted=v{created.version}")
            return run

        experiment = self.registry.create_experiment(
            prompt_name,
            baseline_version=baseline_version,
            challenger_versions=[created.version],
            rollout_percentage=policy.rollout_percentage,
            environment=policy.environment,
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            metadata={
                "source": "self_learning_loop",
                "baseline_version": baseline_version,
                "improvement": run.improvement,
                "guard_allowed": guard.allowed,
                "guard_reasons": guard.reasons,
                "rollback_policy": rollback_policy,
            },
        )
        run.experiment = asdict(experiment)
        scheduled_job = self._ensure_continuous_eval_job(
            agent,
            prompt_name=prompt_name,
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            policy=policy,
            rollback_policy=rollback_policy,
            experiment_id=experiment.id,
            target_version=created.version,
        )
        run.action = SelfLearningLoopAction(
            kind="experiment",
            status="executed",
            reason="created_experiment",
            metadata={
                "experiment_id": experiment.id,
                "version": created.version,
                "continuous_eval_job": scheduled_job,
            },
        )
        run.audit_trail.append(f"experiment={experiment.id}")
        return run

    def _live_metrics(
        self,
        *,
        prompt_name: str,
        prompt_version: str,
        workspace_id: str | None,
        limit: int = 200,
    ) -> dict[str, Any]:
        runs = self.store.list_runs(workspace_id=workspace_id, agent_name=prompt_name, limit=limit)
        relevant = []
        for run in runs:
            metadata = run.metadata or {}
            if metadata.get("prompt_name", prompt_name) != prompt_name:
                continue
            if run.prompt_version != prompt_version:
                continue
            relevant.append(run)
        if not relevant:
            return {
                "avg_latency_ms": 0.0,
                "avg_estimated_cost_usd": 0.0,
                "safety_violation_rate": 0.0,
            }
        latencies = [
            float((run.metadata or {}).get("latency_ms", 0.0) or 0.0)
            for run in relevant
            if (run.metadata or {}).get("latency_ms") is not None
        ]
        safety_violations = 0
        for run in relevant:
            for feedback in self.store.list_feedback(run_id=run.id):
                if "policy_violation" in feedback.failure_types:
                    safety_violations += 1
                    break
        return {
            "avg_latency_ms": round(sum(latencies) / len(latencies), 2) if latencies else 0.0,
            "avg_estimated_cost_usd": 0.0,
            "safety_violation_rate": round(safety_violations / max(1, len(relevant)), 4),
        }

    def _project_candidate_metrics(
        self,
        *,
        baseline_metrics: dict[str, Any],
        baseline_content: str,
        candidate: PromptCandidate,
        patterns: list[FailurePattern],
    ) -> dict[str, Any]:
        baseline_latency = float(baseline_metrics.get("avg_latency_ms", 0.0) or 0.0)
        baseline_cost = float(baseline_metrics.get("avg_estimated_cost_usd", 0.0) or 0.0)
        baseline_safety = float(baseline_metrics.get("safety_violation_rate", 0.0) or 0.0)
        length_ratio = len(candidate.content) / max(1, len(baseline_content))
        latency_ratio = min(1.35, max(0.85, length_ratio))
        cost_ratio = min(1.35, max(0.85, length_ratio))
        if candidate.strategy == "compress":
            latency_ratio = min(latency_ratio, 0.9)
            cost_ratio = min(cost_ratio, 0.9)
        candidate_safety = baseline_safety
        targeted = {failure for pattern in patterns for failure in pattern.failure_types}
        if candidate.strategy in {"patch", "rewrite", "retrieval_grounding"} and targeted & {
            "hallucination",
            "bad_retrieval",
            "policy_violation",
        }:
            candidate_safety = max(0.0, baseline_safety - 0.05)
        return {
            "avg_latency_ms": round(baseline_latency * latency_ratio, 2) if baseline_latency else 0.0,
            "avg_estimated_cost_usd": round(baseline_cost * cost_ratio, 8) if baseline_cost else 0.0,
            "safety_violation_rate": round(candidate_safety, 4),
        }

    def _ensure_continuous_eval_job(
        self,
        agent: Agent,
        *,
        prompt_name: str,
        workspace_id: str | None,
        task_type: str | None,
        user_segment: str | None,
        policy: SelfLearningLoopPolicy,
        rollback_policy: dict[str, Any],
        experiment_id: str | None = None,
        target_version: int | None = None,
    ) -> str | None:
        runner = self.continuous_eval_runner
        if runner is None:
            return None
        job_name = policy.continuous_eval_job_name or (
            f"self-learning-{prompt_name}-{workspace_id or 'global'}-regression"
        )
        if runner.get_job(job_name) is None:
            runner.register_regression_job(
                job_name,
                agent=agent,
                workspace_id=workspace_id,
                agent_name=prompt_name,
                interval_seconds=policy.continuous_eval_interval_seconds,
                limit=policy.continuous_eval_limit,
                metadata={
                    "created_via": "self_learning_loop",
                    "self_learning_policy": {
                        **rollback_policy,
                        "experiment_id": experiment_id,
                        "target_version": target_version,
                    },
                    "task_type": task_type,
                    "user_segment": user_segment,
                },
            )
        return job_name

    def _resolve_reward_scorer(
        self,
        *,
        mode: str,
        workspace_id: str | None,
        prompt_name: str,
        limit: int,
    ) -> RewardScorer:
        return build_reward_scorer(
            mode,
            store=self.store,
            workspace_id=workspace_id,
            agent_name=prompt_name,
            limit=limit,
            fallback_scorer=self.reward_scorer,
        )

    def _historical_reward_score(self, examples: list[Any], scorer: RewardScorer) -> float:
        if not examples:
            return 0.0
        scores = [scorer.score_example(item).score for item in examples]
        return round(sum(scores) / len(scores), 4)

    def _serialize_generation_report(self, report: PromptGenerationReport) -> dict[str, Any]:
        return {
            "prompt_name": report.prompt_name,
            "workspace_id": report.workspace_id,
            "task_type": report.task_type,
            "user_segment": report.user_segment,
            "environment": report.environment,
            "based_on_version": report.based_on_version,
            "pattern_count": report.pattern_count,
            "lesson_count": report.lesson_count,
            "corrected_output_count": report.corrected_output_count,
            "candidates": [
                {
                    "content": candidate.content,
                    "strategy": candidate.strategy,
                    "rationale": candidate.rationale,
                    "metadata": dict(candidate.metadata),
                }
                for candidate in report.candidates
            ],
        }


__all__ = [
    "SelfLearningLoop",
    "SelfLearningLoopAction",
    "SelfLearningLoopPolicy",
    "SelfLearningLoopRun",
]



"""Tests for core data types."""
from logos_adk.types import Message, ToolCall, ToolResult, Usage, Response, StructuredResponse
from pydantic import BaseModel


class MockOutput(BaseModel):
    summary: str
    score: float


def test_message_user():
    msg = Message(role="user", content="hello")
    assert msg.role == "user"
    assert msg.content == "hello"


def test_message_assistant_with_tool_calls():
    tc = ToolCall(id="tc_1", name="search", arguments={"q": "test"})
    msg = Message(role="assistant", content="", tool_calls=[tc])
    assert len(msg.tool_calls) == 1
    assert msg.tool_calls[0].name == "search"


def test_tool_call():
    tc = ToolCall(id="tc_1", name="search", arguments={"query": "ai"})
    assert tc.id == "tc_1"
    assert tc.name == "search"
    assert tc.arguments == {"query": "ai"}


def test_tool_result_success():
    tr = ToolResult(call_id="tc_1", content={"results": [1, 2, 3]})
    assert tr.call_id == "tc_1"
    assert tr.error is None
    assert tr.duration_ms == 0


def test_tool_result_error():
    tr = ToolResult(call_id="tc_1", content=None, error="connection timeout")
    assert tr.error == "connection timeout"


def test_usage():
    u = Usage(input_tokens=100, output_tokens=50)
    assert u.input_tokens == 100
    assert u.output_tokens == 50


def test_response():
    r = Response(
        content="hello",
        tool_calls=[],
        usage=Usage(input_tokens=10, output_tokens=5),
        raw={"id": "msg_123"},
    )
    assert r.content == "hello"
    assert r.tool_calls == []
    assert r.usage.input_tokens == 10
    assert r.raw["id"] == "msg_123"


def test_structured_response():
    parsed = MockOutput(summary="test", score=0.9)
    r = StructuredResponse(
        content='{"summary": "test", "score": 0.9}',
        tool_calls=[],
        usage=Usage(input_tokens=10, output_tokens=20),
        raw={},
        parsed=parsed,
    )
    assert r.parsed.summary == "test"
    assert r.parsed.score == 0.9
    assert isinstance(r, Response)
